/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: spring_embedder.cc 8158 2008-06-10 15:32:07Z gawrilow $"

#include <polymake.h>
#include <random_generators.h>
#include <SpringEmbedder.h>

namespace polymake { namespace graph {

Matrix<double> spring_embedder(const Graph<>& G, perl::OptionSet options)
{
   SpringEmbedder SE(G,options);

   long seed;
   if (!(options["seed"] >> seed)) {
      seed=random_seed();
#ifdef POLY_DEBUG
      cout << "initial random seed=" << seed << endl;
#endif
   }
   UniformRNG< Vector<double> > random_points(3, seed);
   Matrix<double> X(G.nodes(),3);
   SE.start_points(X,random_points.begin());

   int max_iter;
   if (!(options["max-iterations"] >> max_iter)) max_iter=10000;
   if (! SE.calculate(X,random_points,max_iter))
      cerr << "WARNING: spring_embedder not converged after " << max_iter << " iterations" << endl;
   return X;
}

UserFunction4perl("# Produce a 3-d embedding for the graph using the spring embedding algorithm\n"
		  "# along the lines of\n"
		  "#\t Thomas Fruchtermann and Edward Reingold:\n"
		  "#\t Graph Drawing by Force-directed Placement.\n"
		  "#\t Software Practice and Experience Vol. 21, 1129-1164 (1992), no. 11\n"

		  "# The initial node coordinates are chosen randomly on the unit sphere.\n"
		  "# The optional parameter @a seed controls the initial setting.\n"
		  "#\n"
		  "# In the standard setting, the embedding algorithm tries to stretch all edges to the\n"
		  "# same length. If you prefer different edge lengths, pass them via the @a edge_weights option.\n"
		  "#\n"
		  "# If the nodes already have an embedding in `R^d` and there is a linear or abstract objective function defined\n"
		  "# in the coordinate space, it can be used to rearrange the 3-d embedding along the z-axis corresponding\n"
		  "# to the objective function growth.  This mode is enabled with option @a z_ordering.\n"
		  "#\n"
		  "# The embedding algorithm can be fine-tuned with several \"black magic\" options.\n"
		  "# All of them take double values, which are multiplied with internal initial settings,\n"
		  "# so all defaults are equal to 1.\n"
		  "#\n"
		  "# @a scale enlarges the ideal edge length.\n"
		  "# @a balance changes the balance between the edge contraction and node repulsion forces.\n"
		  "# @a inertion and @a viscosity affects how the nodes are moved, and can be used to restrain oscillations.\n"
		  "# @a z-factor changes the relative influence of the objective function on the embedding.\n"
		  "# @a eps controls how far a point may move, to be considered standing still.\n",

		  &spring_embedder,

		  // the next string is *one* long line
		  "spring_embedder(props::Graph<Undirected>, "
		  "   { scale => 1, balance => 1, viscosity => 1, inertion => 1, eps => undef,"
		  "     'z-ordering' => undef, 'z-factor' => undef, 'edge-weights' => undef,"
		  "      seed => undef, 'max-iterations' => 10000 }) ");

} }


// Local Variables:
// c-basic-offset:3
// End:
