#  Copyright (c) 1997-2009
#  Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
#  http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#-----------------------------------------------------------------------------
#  $Project: polymake $$Id: LPparser.pm 9026 2009-01-09 18:03:47Z gawrilow $

package Polymake::polytope::LPparser;
use Polymake::Struct (
   [ new => '$' ],
   [ '$LPfile' => '#1' ],
   '@A', '@B',			# inequalities Ax+B>=0
   '@P', '@Q',			# equations Px+Q=0
   '@C',			# objective function Cx -> min
   '@L', '@U',			# variable bounds  l <= x <= u
   '@X',			# variable index => name
   '%Xi',			# variable name => index
   '$objsense',
);

sub new {
   my $self=&_new;
   open LP, $self->LPfile or die "Can't read ", $self->LPfile, ": $!\n";

   my $lp="";
   local $_;
   while (<LP>) {
      s/\\.*$//;		# comments are starting with '\'
      $lp.=$_;
   }
   close LP;

   $lp =~ /maximize\s|minimize\s|maximum\s|minimum\s|max\s|min\s/is
      or die "no objective function found\n";
   $lp=$';
   $self->objsense= lc(substr($&,0,3)) eq "max" ? "+" : "-";

   $lp =~ /subject\s+to\s|such\s+that\s|s\.?t\.?\s/is
      or die "no constraints found\n";
   $lp=$';
   @{$self->C}=$self->make_vector($`, "+");

   while ($lp and $lp !~ /^\s*bound/is and $lp !~ /^\s*end/is) {
      $lp =~ /[<>]\s*=?|=\s*[<>]?/
         or die "invalid constraint\n";
      $lp=$';
      my ($line, $rel)=($`, $&);
      my @vec=$self->make_vector($line, $rel=~/</ ? "-" : "+");
      my $rhs=parse_number($lp, $rel=~/</ ? "+" : "-");
      die "invalid right hand side in input line '$line'\n" if $rhs eq "" or $rhs eq "-";

      if ($rel=~/[<>]/) {
	 push @{$self->A}, \@vec;
	 push @{$self->B}, $rhs;
      } else {
	 push @{$self->P}, \@vec;
	 push @{$self->Q}, $rhs;
      }
   }

   if ($lp =~ /bounds?\s/is) {
      $lp=$';
      while ($lp and $lp !~ /^\s*end/is) {
         my ($x, $rel, $b, $inf);
         if ($lp =~ /^([^\n]*?)[ \t]+free\s*/i) {
	    $lp=$';
	    my $name=$1;
	    $x=$self->parse_name($name);
            if (defined $x) {
	       undef $self->L->[$x];  undef $self->U->[$x];
	    } else {
	       die "invalid free variable declaration: $name\n";
	    }
            next;
         }

         if ($lp =~ /^([^\n]*?)([<>][ \t]*=?|=[ \t]*[<>]?)([^\n]*?)([<>][ \t]*=?|=[ \t]*[<>]?)/) {
	    $lp="$1$2$3\n$3$4$'";
	 }

         my $xleft=0;
         $x=$self->parse_name($lp);
         if (defined $x) {
	    $xleft=1;
	    $lp =~ /[<>]\s*=?|=\s*[<>]?/
	       or die "invalid bound variable declaration: $lp\n";
	    $lp=$';
	    $rel=$&;
         }

         if ($lp =~ /^\s*([+-])[ \t]*inf(inity)?\s*/) {
	    $inf=$1;
	    $lp=$';
	 } else {
	    $b=parse_number($lp,"+");
	    die "$0: invalid variable bound value\n" if $b eq "" or $b eq "-";
	 }

         if (!$xleft) {
	    $lp =~ /[<>]\s*=?|=\s*[<>]?/
	       or die "invalid bound variable declaration\n";
	    $rel=$&;
            $lp=$';
	    $x=$self->parse_name($lp);
	    die "invalid bound variable declaration\n" unless defined $x;
         }

         if (defined $inf) {
	    die "invalid infinite bound\n"
	       if $inf eq "+" && $xleft != $rel=~/</
	       or $inf eq "-" && $xleft != $rel=~/>/;
	    if ($inf eq "+") {
	       undef $self->U->[$x];
	    } else {
	       undef $self->L->[$x];
	    }
         } else {
	    if ($rel!=/[<>]/) {
	       $self->L->[$x]=$self->U->[$x]=$b;
	    } elsif ($xleft == $rel=~/</) {
	       $self->U->[$x]=$b;
	    } else {
	       $self->L->[$x]=$b;
	    }
         }
      }
   }

   die "unrecognized section\n" if $lp !~ /^\s*end\s*$/is;

   for (my $x=0; $x<=$#{$self->X}; $x++) {
      if (defined $self->L->[$x] and defined $self->U->[$x] and $self->L->[$x]==$self->U->[$x]) {
         my @vec;
         $vec[$x]=-1;
         push @{$self->P}, \@vec;
         push @{$self->Q}, $self->L->[$x];
         next;
      }
      if (defined $self->L->[$x]) {
         my @vec;
         $vec[$x]=1;
         push @{$self->A}, \@vec;
         push @{$self->B}, -$self->L->[$x];
      }
      if (defined $self->U->[$x]) {
         my @vec;
         $vec[$x]=-1;
         push @{$self->A}, \@vec;
         push @{$self->B}, $self->U->[$x];
      }
   }

   my @xorder=sort { $self->X->[$a] cmp $self->X->[$b] } 0..$#{$self->X};

   foreach my $line (@{$self->A}, @{$self->P}, $self->C) {
      @$line=map { $line->[$xorder[$_]]+0 } 0..$#{$self->X};
   }
   @{$self->X}=@{$self->X}[@xorder];

   $self;
}

sub Ineq {
   my $self=shift;
   [ map { [ $self->B->[$_], @{$self->A->[$_]} ] } 0..$#{$self->A} ]
}

sub Eq {
   my $self=shift;
   [ map { [ $self->Q->[$_], @{$self->P->[$_]} ] } 0..$#{$self->P} ]
}

sub Obj {
   my $self=shift;
   [ 0, @{$self->C} ]
}

sub parse_number {		# "line", positive_sign => "advanced line", number
   my ($line,$positive)=@_;
   $line =~ m'^\s*([+-])?([ \t\d]*\.?[ \t\d]*(e[ \t]*[+-]?[ \t\d]+)?)?\s*'si;
   $_[0]=$';
   my ($sign, $number)=($1, lc($2));
   $number=~s/\s//g;                                    # in exact powers of 10
   $number="1$number" if substr($number,0,1) eq "e";    #   the leading 1 can be omitted
   $sign= ($sign or "+") eq $positive ? "" : "-";
   "$sign$number"
}

sub varindex {			# "name" => index
   my ($self, $name)=@_;
   if (exists $self->Xi->{$name}) {
      $self->Xi->{$name}
   } else {
      push @{$self->X}, $name;
      push @{$self->L}, 0;
      $self->Xi->{$name}=$#{$self->X}
   }
}

sub parse_name {		# "line" => index or undef
   my ($self, $line)=@_;
   if ($line =~ m'^\s*([A-Za-z!"#$%&()/,;?@_\'`{}|~][A-Za-z0-9.!"#$%&()/,;?@_\'`{}|~ \t]*)\s*') {
      $_[1]=$';
      my $name=$&;
      $name=~s/\s//g;
      $self->varindex($name)
   } else {
      undef
   }
}

sub make_vector {		# "linear expression", positive_sign => (coefficients)
   my @vec;
   my ($self, $line, $positive)=@_;
   my $iline=$line;
   my ($coef, $i);

   $line =~ s/^.*://;		# remove a possible label

   while ($line) {
      $coef=parse_number($line, $positive);
      $coef.="1" if $coef eq "" or $coef eq "-";
      $i=$self->parse_name($line);
      die "$0: invalid expression in input line '$iline'" unless defined $i;
      $vec[$i]=$coef;
   }
   @vec
}


1

# Local Variables:
# mode: perl
# c-basic-offset:3
# End:
