/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: binary_markov_graph.cc 7879 2008-01-15 19:05:35Z joswig $"

#include <polymake.h>
#include <Rational.h>
#include <Graph.h>
#include <IncidenceMatrix.h>
#include <Array.h>
#include <Vector.h>
#include <Matrix.h>
#include <linalg.h>

namespace polymake { namespace polytope {

   template <typename Container>
   perl::Object binary_markov_graph(const Container& observation)
   {
      perl::Object p("PropagatedPolytope");
      p.set_description() << "Propagated polytope defined by the (maybe) simplest possible (non-trivial) Hidden Markov Model; "
			  << "see Joswig: Polytope Propagation on Graphs, Example 6.9." << endl;

      const Vector<Rational> zero=zero_vector<Rational>(2);
      const Vector<Rational> x1=unit_vector<Rational>(2,0);
      const Vector<Rational> y1=unit_vector<Rational>(2,1);
      Vector<Rational> x2(2),y2(2); x2[0]=y2[1]=2;

      const int n=observation.size();
      Graph<Directed> G(2*n+2); // unique source=0, sink=2*n+1
      EdgeMap<Directed, Vector<Rational> > Trans(G);

      typename Container::const_iterator ob=observation.begin();
      if (*ob) {
	 Trans[G.edge(0,1)]=zero;
	 Trans[G.edge(0,2)]=y1;
      } else {
	 Trans[G.edge(0,1)]=x1;
	 Trans[G.edge(0,2)]=zero;
      }
      
      for (int i=1; i<n; ++i)
	 if (*(++ob)) {
	    Trans[G.edge(2*i-1,2*i+1)]=x1;
	    Trans[G.edge(2*i-1,2*i+2)]=y1;
	    Trans[G.edge(2*i,2*i+1)]=zero;
	    Trans[G.edge(2*i,2*i+2)]=y2;
	 } else {
	    Trans[G.edge(2*i-1,2*i+1)]=x2;
	    Trans[G.edge(2*i-1,2*i+2)]=zero;
	    Trans[G.edge(2*i,2*i+1)]=x1;
	    Trans[G.edge(2*i,2*i+2)]=y1;
	 }

      Trans[G.edge(2*n-1,2*n+1)]=zero;
      Trans[G.edge(2*n,2*n+1)]=zero;

      p.take("SUM_PRODUCT_GRAPH.ADJACENCY") << G;
      p.take("SUM_PRODUCT_GRAPH.TRANSLATIONS") << Trans;
      p.take("AMBIENT_DIM") << 2;
      return p;
   }

   perl::Object binary_markov_graph(const std::string& observation)
   {
      return binary_markov_graph(attach_operation(observation, constant('1'), polymake::operations::eq()));
   }

   UserFunctionTemplate4perl("# Category: Polytope Propagation\n"
			     "# Defines a very simple graph for a polytope propagation related to a Hidden Markov Model.\n"
			     "# The propagated polytope is always a polygon.\n"
			     "# For a detailed description see\n"
			     "#\t M. Joswig: Polytope propagation, in: Algebraic statistics and computational biology\n"
			     "#\t by L. Pachter and B. Sturmfels (eds.), Cambridge, 2005.\n"
			     "# args: Array<Bool> | \"bitstring\"\n",
			     "binary_markov_graph(*)");
} }
