/* Copyright (c) 1997-2007
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: cyclic_caratheodory.cc 7858 2008-01-02 22:33:38Z gawrilow $"

#include <cmath>
#include <Matrix.h>
#include <polymake.h>

namespace polymake { namespace polytope {

perl::Object cyclic_caratheodory(int d, int n)
{
   if ((d < 2) || (d >= n)) {
      throw std::runtime_error("cyclic_caratheodory: (d >= 2) && (n > d)\n");
   }
   if (d % 2) {
      throw std::runtime_error("cyclic_caratheodory: even dimension required.\n");
   }
   
   perl::Object p("Polytope<Rational>");
   p.set_description() << "Cyclic " << d << "-polytope on " << n << " vertices on the trigonometric moment curve" << endl;

   Matrix<double> Vertices(n,d+1);
   double* v=concat_rows(Vertices).begin();

   for (int x = 0; x < n; ++x) {
      *v++ = 1;
      for (int i = 1; i <= d/2; ++i) {
	 *v++ = cos(i*x);
	 *v++ = sin(i*x);
      }
   }

   p.take("AMBIENT_DIM") << d;
   p.take("DIM") << d;
   p.take("N_VERTICES") << n;
   p.take("VERTICES") << Vertices;
   p.take("BOUNDED") << true;
   return p;
}

UserFunction4perl("# Category: Producing from scratch\n"
		  "# Produce a @a d-dimensional cyclic polytope with @a n points.\n"
		  "# Prototypical example of a neighborly polytope.  Combinatorics completely known\n"
		  "# due to Gale's evenness criterion.  Coordinates are chosen on the trigonometric\n"
		  "# moment curve.\n"
		  "# args: d, n\n",
		  &cyclic_caratheodory,"cyclic_caratheodory($$)");

} }


