/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: polarize.cc 9054 2009-02-08 10:55:41Z joswig $"

#include <polymake.h>
#include <Matrix.h>
#include <Array.h>
#include <IncidenceMatrix.h>
#include <linalg.h>

namespace polymake { namespace polytope {

template <typename Scalar>
perl::Object polarize(perl::Object p_in, perl::OptionSet options)
{
   bool some_section_read = false;
   const bool no_coordinates=options["noc"];

   perl::Object p_out("Polytope<Rational>");
   p_out.set_description() << "Polytope polarized from " << p_in.name() << endl;

   if (no_coordinates || p_in.exists("VERTICES_IN_FACETS")) {
      const IncidenceMatrix<> VIF=p_in.give("VERTICES_IN_FACETS");
      p_out.take("VERTICES_IN_FACETS") << T(VIF);
      some_section_read = true;
   }

   if (!no_coordinates) {
      const bool is_bounded=p_in.give("BOUNDED"),
	         is_centered=p_in.give("CENTERED");
      const int  ambient_dim=p_in.give("AMBIENT_DIM"),
	         dim=p_in.give("DIM");
      if (!is_bounded || !is_centered || dim!=ambient_dim)
	 throw std::runtime_error("polarize: polytope must be bounded, centered, and full-dimensional");

      Matrix<Scalar> ineq, AH;
      std::string got_property;
      if (p_in.lookup_with_property_name("VERTICES | POINTS", got_property) >> ineq) {
	 ineq.minor(All, range(1,ineq.cols()-1)).negate();
	 p_out.take( got_property=="VERTICES" ? "FACETS" : "INEQUALITIES") << ineq;
	 p_out.take("AFFINE_HULL") << AH;
	 some_section_read = true;
      }

      Matrix<Scalar> pts;
      if (p_in.lookup_with_property_name("FACETS | INEQUALITIES", got_property) >> pts) {
	 pts.minor(All, range(1,pts.cols()-1)).negate();
	 p_out.take( got_property=="FACETS" ? "VERTICES" : "POINTS") << pts;
	 some_section_read = true;
      }

      // let's see if we produced something at all
      if (!some_section_read)
	 throw std::runtime_error("polarize: no useful coordinate section found in input");

      // if p was obtained by some transformation,
      // then the reverse transformation also makes sense for the polar
      Matrix<Scalar> tau;
      if (p_in.get_attachment("REVERSE_TRANSFORMATION") >> tau)
	 p_out.attach("REVERSE_TRANSFORMATION") << T(inv(tau));
	
      // now let's claim that the result is bounded, centered, and full-dimensional
      p_out.take("BOUNDED") << true;
      p_out.take("CENTERED") << true;
      p_out.take("AMBIENT_DIM") << ambient_dim;
      p_out.take("DIM") << dim;
   }

   Array<std::string> labels;
   if (p_in.lookup("VERTEX_LABELS") >> labels)
      p_out.take("FACET_LABELS") << labels;
   if (p_in.lookup("FACET_LABELS") >> labels)
      p_out.take("VERTEX_LABELS") << labels;

   return p_out;
}

UserFunctionTemplate4perl("# Category: Transforming a polyhedron\n"
			  "# Given a bounded, centered, and full-dimensional polytope, produce its polar,\n"
			  "# that is, polar with respect to the standard Euclidean scalar product.\n"
			  "# If the option @a noc (no coordinates) is set, then only combinatorial information is handled.\n",
			  "polarize<Scalar> (Polytope<Scalar>, { noc => 0 })");
} }
