/* Copyright (c) 1997-2004
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Version: 1.3.1 $$Id: pyramid.cc 7861 2008-01-03 23:13:27Z gawrilow $"

#include <vector>
#include <Rational.h>
#include <Matrix.h>
#include <Vector.h>
#include <IncidenceMatrix.h>
#include <polymake.h>

namespace polymake { namespace polytope {

   perl::Object pyramid(perl::Object p_in, const Rational& z, perl::OptionSet options)
   {
      const bool noc = options["noc"],
             relabel = options["relabel"];

      if (!z && !noc)
	 throw std::runtime_error("pyramid: z must be non-zero");

      int n_vertices=0;
      perl::Object p_out("Polytope<Rational>");
      p_out.set_description() << "pyramid over " << p_in.name() << endl;

      if (noc || p_in.exists("VERTICES_IN_FACETS")) {
	 const IncidenceMatrix<> VIF=p_in.give("VERTICES_IN_FACETS");
	 n_vertices=VIF.cols();
	 const int n_facets=VIF.rows();
	 const IncidenceMatrix<> VIF_out= (VIF | sequence(0,n_facets))	// original vertices + the new top vertex
                                          / sequence(0, n_vertices);	// original polytope becomes the bottom facet

	 p_out.take("N_VERTICES") << n_vertices+1;
	 p_out.take("VERTICES_IN_FACETS") << VIF_out;
      }
      if (noc) {
	 if (p_in.exists("DIM")) {
	    const int dim=p_in.give("DIM");
	    p_out.take("DIM") << dim+1;
	 }
      } else {
	 const Matrix<Rational> V=p_in.give("VERTICES");
	 const Vector<Rational> z0=p_in.give("REL_INT_POINT");
	 n_vertices=V.rows();
	 p_out.take("VERTICES") << (V | zero_vector<Rational>(n_vertices)) /
	                           (z0 | z);
      }

      if (relabel) {
	 std::vector<std::string> labels(n_vertices+1);
	 read_labels(p_in, "VERTEX_LABELS", labels);
	 labels[n_vertices]="Apex";
	 p_out.take("VERTEX_LABELS") << labels;
      }

      return p_out;
   }

   UserFunction4perl("# Category: Producing a new polyhedron from others\n"
		     "# Make a pyramid over a polyhedron.\n"
		     "# The pyramid is the convex hull of the input polyhedron $P$ and a point $v$\n"
		     "# outside the affine span of $P$. For bounded polyhedra, the projection of $v$\n"
		     "# to the affine span of $P$ coincides with the vertex barycenter of $P$.\n"
		     "# \n"
		     "# @a z is the distance between the vertex barycenter and $v$, default value is $1$.\n"
		     "# \n"
		     "# If the option @a noc (no coordinates) is set, then only combinatorial information is produced.\n"
		     "# The option @a relabel creates an additional section @see VERTEX_LABELS.\n"
		     "# The vertices from the original polytope keep their labels;\n"
		     "# the new top vertex is labeled \"Apex\".\n"
		     "# args: [ z ], [noc => Bool], [relabel => Bool]\n",
		     &pyramid, "pyramid(Polytope; $=1, { noc => undef, relabel => undef })");
} }
