/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ident "$Project: polymake $$Id: steiner_points.cc 8921 2008-11-17 12:29:14Z gawrilow $"

#include <cmath>
#include <polymake.h>
#include <Matrix.h>
#include <Vector.h>
#include <Graph.h>
#include <Rational.h>
#include <random_generators.h>
#include <HasseDiagram.h>

namespace polymake { namespace polytope {

namespace {

   template <typename Coord> inline
   Coord calc_weight(int a, int b, Coord*) { return Coord(a)/Coord(b); }

   inline
   Rational calc_weight(int a, int b, Rational*) { return Rational(a,b); }

   template <typename Coord>
   Coord calc_weight(const Matrix<double>& Vd, Matrix<double> adjVert,
		     const int i, const int dim, const int nop,
		     UniformRNG< Vector<double> >& random_source,
		     const double eps)
   {
      // get the perpendicular vectors for actual vertex
      for (  Entire< Rows< Matrix<double> > >::iterator ri = entire(rows(adjVert)); !ri.at_end(); ++ri )
	 *ri = Vd[i] - *ri;

      int countwhile = 0;
      int old_out = 0;
      int out = 0;
      double step_dist;

      // check more points on the sphere until weights "converge"
      do {
	 //reinitialize some values
	 step_dist = 0;
	 countwhile++;
	 Matrix<double> rand_points((countwhile == 1)?50*nop:nop, dim+1);
	 rand_points.col(0).fill(1);
	 std::copy(random_source.begin(), entire(rows(rand_points.minor(All, range(1,dim)).top())));

	 // check how many points are in- and outside the cone
	 old_out = out;
	 for ( Entire< Rows< Matrix<double> > >::iterator ri1=entire(rows(rand_points)); !ri1.at_end(); ++ri1 )
	    for ( Entire< Rows< Matrix<double> > >::iterator ri2= entire(rows(adjVert)); !ri2.at_end(); ++ri2 )
	       if ( (*ri1)*(*ri2) < 0) {
		  out++;
		  break;
	       }

	 step_dist = (countwhile!=1)?(double(out)/countwhile)/(double(old_out)/(countwhile-1)):2;

      } while (abs(step_dist-1) >= eps);

      return calc_weight((countwhile+49)*nop-out, (countwhile+49)*nop, (Coord*)0);
   }
}

   template <typename Coord>
   Matrix<Coord> all_steiner_points(perl::Object p, perl::OptionSet options)
   {
      const int dim=p.give("DIM");
      const Matrix<Coord> V=p.give("VERTICES");
      const Matrix<double> Vd(V);
      const Graph<> G=p.give("GRAPH.ADJACENCY");
      const HasseDiagram HDiagram=p.give("HASSE_DIAGRAM");

      long seed;
      if (!(options["seed"] >> seed)) seed=random_seed();
      UniformRNG< Vector<double> > random_source(dim,seed);

      const double eps=options["eps"];

      const int n_stp = HDiagram.nodes() - HDiagram.node_range_of_dim(1).size() - HDiagram.node_range_of_dim(0).size() - 1;
      Matrix<Coord> steiner_points(n_stp,dim+1);
      typename Rows< Matrix<Coord> >::iterator si=rows(steiner_points).begin();

      for (int d = dim; d >= 2; --d) {
	 const int nop = 10*(1<<d);

	 //iterate over the faces of current dimension d
	 for (Entire<HasseDiagram::nodes_of_dim_set>::iterator fi=entire(HDiagram.node_range_of_dim(d)); !fi.at_end(); ++fi, ++si) {

	    const Set<int>& face = HDiagram.face(*fi);
	    Vector<Coord> weights(face.size());
	    Coord sw(0);
	    typename Vector<Coord>::iterator wi=weights.begin();

	    //iterate over the vertices of the face
	    for ( Entire< Set<int> >::const_iterator vi=entire(face); !vi.at_end(); ++vi, ++wi ) {
	       const Coord w=calc_weight<Coord>(Vd, Vd.minor(G.adjacent_nodes(*vi)*face, All),
						*vi, dim, nop, random_source, eps);
	       sw+=w;
	       *wi=w;
	    }

	    weights /= sw;
	    *si = weights*V.minor(face,All);
	 }
      }

      return steiner_points;
   }

   template <typename Coord>
   Vector<Coord> steiner_point(perl::Object p, perl::OptionSet options)
   {
      const int dim=p.give("DIM");
      const Matrix<Coord> V=p.give("VERTICES");
      const Matrix<double> Vd(V);
      const Graph<> G=p.give("GRAPH.ADJACENCY");

      long seed;
      if (!(options["seed"] >> seed)) seed=random_seed();
      UniformRNG< Vector<double> > random_source(dim,seed);

      const double eps=options["eps"];

      const int nop = 10*(1<<dim);

      Vector<Coord> weights(V.rows());
      Coord sw(0);
      typename Vector<Coord>::iterator wi=weights.begin();

      for (Entire< Nodes< Graph<> > >::const_iterator n=entire(nodes(G)); !n.at_end(); ++n, ++wi) {
	 const Coord w=calc_weight<Coord>(Vd, Vd.minor(n.adjacent_nodes(),All),
					  *n, dim, nop, random_source, eps);
	 sw+=w;
	 *wi=w;
      }

      weights/=sw;
      return weights*V;
   }

   UserFunctionTemplate4perl("# Category: Other\n"
			     "# Compute the Steiner points of all faces of a polytope using a\n"
			     "# randomized approximation of the angles.\n\n"
			     "# Option @c eps controls the accuracy of the angles computed.\n"
			     "# Option @c seed may be set to initialize the random calculation of the angles.\n\n"
			     "# Polytope must be bounded.\n"
			     "# @author Thilo Rörig <thilosch@math.tu-berlin.de>\n",
			     "all_steiner_points<Coord>(Polytope<Coord> { seed => undef, eps => 0.1 })");

   UserFunctionTemplate4perl("# Category: Other\n"
			     "# Compute the Steiner point of a polytope using a randomized approximation of the angles.\n\n"
			     "# Option @c eps controls the accuracy of the angles computed.\n"
			     "# Option @c seed may be set to initialize the random calculation of the angles.\n\n"
			     "# @author Thilo Rörig <thilosch@math.tu-berlin.de>\n",
			     "steiner_point<Coord>(Polytope<Coord> { seed => undef, eps => 0.1 })");
} }
