/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_INCIDENCE_MATRIX_H
#define _POLYMAKE_INCIDENCE_MATRIX_H "$Project: polymake $$Id: IncidenceMatrix.h 8867 2008-10-27 23:11:37Z gawrilow $"

#ifndef _POLYMAKE_GENERIC_INCIDENCE_MATRIX_H
#  include <GenericIncidenceMatrix.h>
#endif
#ifndef _POLYMAKE_SPARSE2D_H
#  include <sparse2d.h>
#endif
#ifndef _POLYMAKE_SET_H
#  include <Set.h>
#endif
#ifndef _POLYMAKE_PERMUTATIONS_H
#  include <permutations.h>
#endif

namespace pm {

template <typename Set>
class incidence_proxy_base {
protected:
   Set* s;
   int j;

   bool get() const { return s->exists(j); }

   void insert() { s->insert(j); }

   void erase() { s->erase(j); }

   void toggle() { s->toggle(j); }
public:
   typedef bool value_type;

   incidence_proxy_base(Set& s_arg, int j_arg)
      : s(&s_arg), j(j_arg) {}
};

template <typename Tree>
class incidence_line
   : public modified_tree< incidence_line<Tree>,
			   list( Hidden< sparse2d::line<Tree> >,
				 Operation< BuildUnaryIt<operations::index2element> > ) >,
     public GenericMutableSet<incidence_line<Tree>, int, operations::cmp> {
protected:
   ~incidence_line();
public:
   incidence_line& operator= (const incidence_line& l)
   {
      return incidence_line::generic_mutable_type::operator=(l);
   }
   using incidence_line::generic_mutable_type::operator=;

   template <class _Set>
   incidence_line& operator= (const Complement<_Set, int, operations::cmp>& cl)
   {
      return incidence_line::generic_mutable_type::operator=(sequence(0, this->dim()) * cl);
   }

   bool contains(int e) const { return this->exists(e); }
};

template <typename Tree>
struct check_container_feature<incidence_line<Tree>, sparse_compatible> : True {};

template <sparse2d::restriction_kind restriction=sparse2d::only_rows>
class RestrictedIncidenceMatrix
   : public matrix_methods<RestrictedIncidenceMatrix<restriction>, bool> {
protected:
   typedef sparse2d::Table<nothing, false, restriction> table_type;
   table_type data;

   template <typename Iterator>
   void _copy(Iterator src, True)
   {
      std::copy(src, entire(pm::rows(*this)));
   }
   template <typename Iterator>
   void _copy(Iterator src, False)
   {
      std::copy(src, entire(pm::cols(*this)));
   }

   typedef incidence_proxy_base<incidence_line<typename table_type::primary_tree_type> > proxy_base;

public:
   typedef bool value_type;
   typedef sparse_elem_proxy<proxy_base> reference;
   typedef const bool const_reference;
   static const bool prefer_concat=false;

   RestrictedIncidenceMatrix() : data(nothing()) {}
   explicit RestrictedIncidenceMatrix(int n) : data(n) {}

   template <typename Iterator>
   RestrictedIncidenceMatrix(int n, Iterator src)
      : data(n)
   {
      _copy(src, bool2type<restriction==sparse2d::only_rows>());
   }

   template <typename Container>
   RestrictedIncidenceMatrix(const Container& src,
			     typename enable_if<nothing*, isomorphic_to_container_of<Container, Set<int> >::value &&
			                                  restriction==sparse2d::only_rows>::type=0)
      : data(src.size())
   {
      _copy(src.begin(), True());
   }

   void swap(RestrictedIncidenceMatrix& M) { data.swap(M.data); }

   void clear() { data.clear(); }

protected:
   proxy_base _random(int i, int j, False)
   {
      return proxy_base(this->row(i), j);
   }
   proxy_base _random(int i, int j, True)
   {
      return proxy_base(this->col(j), i);
   }
   bool _random(int i, int j, False) const
   {
      return this->row(i).exists(j);
   }
   bool _random(int i, int j, True) const
   {
      return this->col(j).exists(i);
   }
public:
   reference operator() (int i, int j)
   {
      return _random(i, j, bool2type<restriction==sparse2d::only_cols>());
   }
   const_reference operator() (int i, int j) const
   {
      return _random(i, j, bool2type<restriction==sparse2d::only_cols>());
   }

   bool exists(int i, int j) const
   {
      return _random(i, j, bool2type<restriction==sparse2d::only_cols>());
   }

private:
   template <typename RowCol, typename _Set>
   void _append(RowCol& row_col, const _Set& set, int i)
   {
      for (typename Entire<_Set>::const_iterator s=entire(set); !s.at_end(); ++s)
	 row_col[*s].push_back(i);
   }

   template <typename Iterator>
   void _append_rows(int n, Iterator src, True)
   {
      int oldrows=data.rows();
      data.resize_rows(oldrows+n);
      std::copy(src, pm::rows(*this).begin()+oldrows);
   }

   template <typename Iterator>
   void _append_rows(int, Iterator src, False)
   {
      for (int r=data.rows(); !src.at_end(); ++src, ++r)
	 _append(pm::cols(*this), *src, r);
   }

   template <typename Iterator>
   void _append_cols(int n, Iterator src, True)
   {
      int oldcols=data.cols();
      data.resize_cols(oldcols+n);
      std::copy(src, pm::cols(*this).begin()+oldcols);
   }

   template <typename Iterator>
   void _append_cols(int, Iterator src, False)
   {
      for (int c=data.cols(); !src.at_end(); ++src, ++c)
	 _append(pm::rows(*this), *src, c);
   }

public:
   template <typename _Matrix>
   RestrictedIncidenceMatrix& operator/= (const GenericIncidenceMatrix<_Matrix>& m)
   {
      _append_rows(m.rows(), entire(pm::rows(m)), bool2type<restriction==sparse2d::only_rows>());
      return *this;
   }

   template <typename _Set>
   RestrictedIncidenceMatrix& operator/= (const GenericSet<_Set, int, operations::cmp>& s)
   {
      _append_rows(1, entire(item2container(s.top())), bool2type<restriction==sparse2d::only_rows>());
      return *this;
   }

   template <typename _Matrix>
   RestrictedIncidenceMatrix& operator|= (const GenericIncidenceMatrix<_Matrix>& m)
   {
      _append_cols(m.cols(), entire(pm::cols(m)), bool2type<restriction==sparse2d::only_cols>());
      return *this;
   }

   template <typename _Set>
   RestrictedIncidenceMatrix& operator|= (const GenericSet<_Set, int, operations::cmp>& s)
   {
      _append_cols(1, entire(item2container(s.top())), bool2type<restriction==sparse2d::only_cols>());
      return *this;
   }

   void squeeze() { data.squeeze(); }

   template <typename Iterator>
   void permute_rows(Iterator perm)
   {
      data.permute_rows(perm, False());
   }

   template <typename Iterator>
   void permute_cols(Iterator perm)
   {
      data.permute_cols(perm, False());
   }

   template <typename Iterator>
   void permute_inv_rows(Iterator inv_perm)
   {
      data.permute_rows(inv_perm, True());
   }

   template <typename Iterator>
   void permute_inv_cols(Iterator inv_perm)
   {
      data.permute_cols(inv_perm, True());
   }

#ifdef POLY_DEBUG
   void check() const { data.check(); }
#endif
   friend class Rows<RestrictedIncidenceMatrix>;
   friend class Cols<RestrictedIncidenceMatrix>;
   template <typename> friend class IncidenceMatrix;
};

template <sparse2d::restriction_kind restriction>
class Rows< RestrictedIncidenceMatrix<restriction> >
   : public redirected_container< Rows< RestrictedIncidenceMatrix<restriction> >,
				  list( Container< sparse2d::Rows<nothing, false, restriction, operations::masquerade<incidence_line> > >,
					MasqueradedTop ) > {
   typedef redirected_container<Rows> _super;
public:
   typename _super::container& get_container()
   {
      return reinterpret_cast<typename _super::container&>(this->hidden().data);
   }
   const typename _super::container& get_container() const
   {
      return reinterpret_cast<const typename _super::container&>(this->hidden().data);
   }

   int size() const
   {
      return this->hidden().data.rows();
   }
   bool empty() const
   {
      return !size();
   }
   void resize(int n)
   {
      this->hidden().data.resize_rows(n);
   }
};

template <sparse2d::restriction_kind restriction>
class Cols< RestrictedIncidenceMatrix<restriction> >
   : public redirected_container< Cols< RestrictedIncidenceMatrix<restriction> >,
				  list( Container< sparse2d::Cols<nothing, false, restriction, operations::masquerade<incidence_line> > >,
					MasqueradedTop ) > {
   typedef redirected_container<Cols> _super;
public:
   typename _super::container& get_container()
   {
      return reinterpret_cast<typename _super::container&>(this->hidden().data);
   }
   const typename _super::container& get_container() const
   {
      return reinterpret_cast<const typename _super::container&>(this->hidden().data);
   }

   int size() const
   {
      return this->hidden().data.cols();
   }
   bool empty() const {
      return !size();
   }
   void resize(int n)
   {
      this->hidden().data.resize_cols(n);
   }
};

template <sparse2d::restriction_kind restriction>
struct spec_object_traits< RestrictedIncidenceMatrix<restriction> >
   : spec_object_traits<is_container> {
   static const int dimension=2;

   typedef typename if_else<restriction==sparse2d::only_rows,
			    Rows< RestrictedIncidenceMatrix<restriction> >,
			    Cols< RestrictedIncidenceMatrix<restriction> > >::type serialized;

   static serialized& serialize(RestrictedIncidenceMatrix<restriction>& M)
   {
      return reinterpret_cast<serialized&>(M);
   }
   static const serialized& serialize(const RestrictedIncidenceMatrix<restriction>& M)
   {
      return reinterpret_cast<const serialized&>(M);
   }
};

template <typename symmetric>
class IncidenceMatrix
   : public GenericIncidenceMatrix< IncidenceMatrix<symmetric> > {
protected:
   typedef sparse2d::Table<nothing,symmetric::value> table_type;
   shared_object<table_type, AliasHandler<shared_alias_handler> > data;

   friend IncidenceMatrix& make_mutable_alias(IncidenceMatrix& alias, IncidenceMatrix& owner)
   {
      alias.data.make_mutable_alias(owner.data);
      return alias;
   }

   // initialize from a dense boolean sequence in row order
   template <typename Iterator>
   void _init(Iterator src, True)
   {
      const int n=this->cols();
      for (typename Entire< Rows<IncidenceMatrix> >::iterator r_i=entire(pm::rows(*this)); !r_i.at_end(); ++r_i) {
	 int i=0;
	 if (symmetric::value) {
	    i=r_i.index();
	    std::advance(src,i);
	 }
	 for (; i<n; ++i, ++src)
	    if (*src) r_i->push_back(i);
      }
   }

   // initialize rowwise from a sequence of sets
   template <typename Iterator>
   void _init(Iterator src, False)
   {
      std::copy(src, entire(pm::rows(*this)));
   }

   typedef incidence_proxy_base< incidence_line<typename table_type::primary_tree_type> > proxy_base;

public:
   typedef sparse_elem_proxy<proxy_base> reference;
   typedef const bool const_reference;

   IncidenceMatrix() {}

   IncidenceMatrix(int m, int n)
      : data( make_constructor(m,n,(table_type*)0) ) {}

   template <typename Iterator>
   IncidenceMatrix(int m, int n, Iterator src)
      : data( make_constructor(m,n,(table_type*)0) )
   {
      _init(src, bool2type<(object_traits<typename iterator_traits<Iterator>::value_type>::total_dimension==0)>());
   }

   IncidenceMatrix(const GenericIncidenceMatrix<IncidenceMatrix>& M)
      : data(M.top().data) {}

   template <typename Matrix2>
   IncidenceMatrix(const GenericIncidenceMatrix<Matrix2>& M,
		   typename enable_if<nothing*, !symmetric::value || Matrix2::is_symmetric>::type=0)
      : data( make_constructor(M.rows(), M.cols(), (table_type*)0) )
   {
      _init(pm::rows(M).begin(), False());
   }

   template <sparse2d::restriction_kind restriction>
   explicit IncidenceMatrix(RestrictedIncidenceMatrix<restriction>& M,
			    typename disable_if<typename cons<nothing*, RestrictedIncidenceMatrix<restriction> >::head, symmetric::value>::type=0)
      : data( make_constructor(M.data,(table_type*)0) ) {}

   // number of columns not known in advance
   template <typename Container>
   IncidenceMatrix(const Container& src,
		   typename enable_if<nothing*, isomorphic_to_container_of<Container, Set<int> >::value &&
			                        !symmetric::value>::type=0)
   {
      RestrictedIncidenceMatrix<> Mt(src);
      *this=Mt;
   }

   // number of columns given explicitly
   template <typename Container>
   IncidenceMatrix(const Container& src,
		   typename enable_if<int, isomorphic_to_container_of<Container, Set<int> >::value &&
			                   !symmetric::value>::type c)
      : data( make_constructor(src.size(), c, (table_type*)0) )
   {
      _init(src.begin(), False());
   }

private:
   template <typename Input>
   void input(GenericInput<Input>& is, False)
   {
      RestrictedIncidenceMatrix<> Mt;
      is.top() >> Mt;
      *this=Mt;
   }

   template <typename Input>
   void input(GenericInput<Input>& is, True)
   {
      is.top() >> pm::rows(*this);
   }

   template <typename Input>
   void input(GenericInput<Input>& is)
   {
      input(is, symmetric());
   }
public:
   template <typename Input> friend
   Input& operator>> (GenericInput<Input>& is, IncidenceMatrix& M)
   {
      M.input(is);
      return is.top();
   }

   IncidenceMatrix& operator= (const IncidenceMatrix& M)
   {
      assign(M);
      return *this;
   }

   using IncidenceMatrix::generic_type::operator=;

   template <sparse2d::restriction_kind restriction>
   typename disable_if<typename cons<IncidenceMatrix, RestrictedIncidenceMatrix<restriction> >::head, symmetric::value>::type&
   operator= (RestrictedIncidenceMatrix<restriction>& M)
   {
      data=make_constructor(M.data,(table_type*)0);
      return *this;
   }

   void swap(IncidenceMatrix& M) { data.swap(M.data); }

   friend void relocate(IncidenceMatrix* from, IncidenceMatrix* to)
   {
      relocate(&from->data, &to->data);
   }

   void resize(int m, int n) { data->resize(m,n); }

   void clear() { data.apply(shared_clear()); }

   void clear(int r, int c) { data.apply(typename table_type::shared_clear(r,c)); }

   reference operator() (int i, int j)
   {
      INDEX_CHECK((i<0 || i>=this->rows() || j<0 || j>= this->cols()), "IncidenceMatrix - element indices out of range");
      return proxy_base(this->row(i),j);
   }

   const_reference operator() (int i, int j) const
   {
      INDEX_CHECK((i<0 || i>=this->rows() || j<0 || j>= this->cols()), "IncidenceMatrix - element indices out of range");
      return exists(i,j);
   }

   bool exists(int i, int j) const
   {
      INDEX_CHECK((i<0 || i>=this->rows() || j<0 || j>= this->cols()), "IncidenceMatrix::exists - element indices out of range");
      return data->row(i).exists(j);
   }

   template <typename Matrix2>
   typename disable_if<typename cons<IncidenceMatrix, Matrix2>::head, symmetric::value>::type&
   operator/= (const GenericIncidenceMatrix<Matrix2>& M)
   {
      const int r=M.rows();
      if (r) {
	 DIMENSION_CHECK((this->cols() != M.cols()), "append_rows - column dimensions mismatch");
	 const int old_rows=this->rows();
	 data.apply(typename table_type::shared_add_rows(r));
	 std::copy(entire(pm::rows(M)), pm::rows(*this)+old_rows);
      }
      return *this;
   }

   template <typename Set2>
   typename disable_if<typename cons<IncidenceMatrix, Set2>::head, symmetric::value>::type&
   operator/= (const GenericSet<Set2, int, operations::cmp>& s)
   {
      DIMENSION_CHECK((!s.top().empty() && s.top().back() >= this->cols()), "append_row - set too long");
      const int old_rows=this->rows();
      data.apply(typename table_type::shared_add_rows(1));
      this->row(old_rows)=s;
      return *this;
   }

   template <typename Set2>
   typename disable_if<typename cons<IncidenceMatrix, Set2>::head, symmetric::value>::type&
   operator/= (const Complement<Set2, int, operations::cmp>& s)
   {
      DIMENSION_CHECK((!s.base().empty() && s.base().back() >= this->cols()), "append_row - set too long");
      const int old_rows=this->rows();
      data.apply(typename table_type::shared_add_rows(1));
      this->row(old_rows)=sequence(0,this->cols()) * s;
      return *this;
   }

   template <typename Matrix2>
   typename disable_if<typename cons<IncidenceMatrix, Matrix2>::head, symmetric::value>::type&
   operator|= (const GenericIncidenceMatrix<Matrix2>& M)
   {
      const int c=M.cols();
      if (c) {
	 DIMENSION_CHECK((this->rows() != M.rows()), "append_cols - row dimensions mismatch");
	 const int old_cols=this->cols();
	 data.apply(typename table_type::shared_add_cols(c));
	 std::copy(entire(pm::cols(M)), pm::cols(*this)+old_cols);
      }
      return *this;
   }

   template <typename Set2>
   typename disable_if<typename cons<IncidenceMatrix, Set2>::head, symmetric::value>::type&
   operator|= (const GenericSet<Set2, int, operations::cmp>& s)
   {
      DIMENSION_CHECK((!s.top().empty() && s.top().back() >= this->rows()), "append_col - set too long");
      const int old_cols=this->cols();
      data.apply(typename table_type::shared_add_cols(1));
      this->col(old_cols)=s;
      return *this;
   }

   template <typename Set2>
   typename disable_if<typename cons<IncidenceMatrix, Set2>::head, symmetric::value>::type&
   operator|= (const Complement<Set2, int, operations::cmp>& s)
   {
      DIMENSION_CHECK((!s.base().empty() && s.base().back() >= this->rows()), "append_col - set too long");
      const int old_cols=this->cols();
      data.apply(typename table_type::shared_add_cols(1));
      this->col(old_cols)=sequence(0,this->rows()) * s;
      return *this;
   }

   template <typename row_number_consumer, typename col_number_consumer>
   void squeeze(row_number_consumer rnc, col_number_consumer cnc) { data->squeeze(rnc,cnc); }

   template <typename row_number_consumer>
   void squeeze(row_number_consumer rnc) { data->squeeze(rnc); }

   void squeeze() { data->squeeze(); }

   template <typename row_number_consumer>
   void squeeze_rows(row_number_consumer rnc) { data->squeeze_rows(rnc); }

   void squeeze_rows() { data->squeeze_rows(); }

   template <typename col_number_consumer>
   void squeeze_cols(col_number_consumer cnc) { data->squeeze_cols(cnc); }

   void squeeze_cols() { data->squeeze_cols(); }

   template <typename Iterator>
   void permute_rows(Iterator perm)
   {
      data->permute_rows(perm, False());
   }

   template <typename Iterator>
   void permute_cols(Iterator perm)
   {
      data->permute_cols(perm, False());
   }

   template <typename Iterator>
   void permute_inv_rows(Iterator inv_perm)
   {
      data->permute_rows(inv_perm, True());
   }

   template <typename Iterator>
   void permute_inv_cols(Iterator inv_perm)
   {
      data->permute_cols(inv_perm, True());
   }

   template <typename Perm, typename InvPerm>
   IncidenceMatrix copy_permuted(const Perm& perm, const InvPerm& inv_perm,
				 typename enable_if<typename cons<nothing*,Perm>::head, symmetric::value>::type=0) const
   {
      const int n=this->rows();
      IncidenceMatrix result(n,n);
      result.data.get()->copy_permuted(*data, perm, inv_perm);
      return result;
   }

#ifdef POLY_DEBUG
   void check() const { data->check(); }
#endif
protected:
   friend class Rows<IncidenceMatrix>;
   friend class Cols<IncidenceMatrix>;
   friend class GenericIncidenceMatrix<IncidenceMatrix>;

   void assign(const GenericIncidenceMatrix<IncidenceMatrix>& M) { data=M.top().data; }

   template <typename Matrix>
   void assign(const GenericIncidenceMatrix<Matrix>& M)
   {
      if (data.is_shared() || this->rows() != M.rows() || this->cols() != M.cols())
	 // circumvent the symmetry checks, they are already done in GenericIncidenceMatrix methods
	 assign(IncidenceMatrix(M.rows(),M.cols(),pm::rows(M).begin()));
      else
	 GenericIncidenceMatrix<IncidenceMatrix>::assign(M);
   }
};

template <typename symmetric>
struct check_container_feature< IncidenceMatrix<symmetric>, Symmetric > : symmetric {};

template <typename symmetric>
class Rows< IncidenceMatrix<symmetric> >
   : public redirected_container< Rows< IncidenceMatrix<symmetric> >,
				  list( Container< sparse2d::Rows<nothing, symmetric::value, sparse2d::full, operations::masquerade<incidence_line> > >,
					MasqueradedTop ) > {
   typedef redirected_container<Rows> _super;
public:
   typename _super::container& get_container()
   {
      return reinterpret_cast<typename _super::container&>(*this->hidden().data);
   }
   const typename _super::container& get_container() const
   {
      return reinterpret_cast<const typename _super::container&>(*this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data->resize_rows(n);
   }
};

template <typename symmetric>
class Cols< IncidenceMatrix<symmetric> >
   : public redirected_container< Cols< IncidenceMatrix<symmetric> >,
				  list( Container< sparse2d::Cols<nothing, symmetric::value, sparse2d::full, operations::masquerade<incidence_line> > >,
					MasqueradedTop ) > {
   typedef redirected_container<Cols> _super;
public:
   typename _super::container& get_container()
   {
      return reinterpret_cast<typename _super::container&>(*this->hidden().data);
   }
   const typename _super::container& get_container() const
   {
      return reinterpret_cast<const typename _super::container&>(*this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data->resize_cols(n);
   }
};

// convolution of two incidence relations
template <typename _Matrix1, typename _Matrix2> inline
IncidenceMatrix<>
convolute(const GenericIncidenceMatrix<_Matrix1>& M1, const GenericIncidenceMatrix<_Matrix2>& M2)
{
   DIMENSION_CHECK((M1.cols() != M2.rows()), "convolute - dimensions mismatch");

   IncidenceMatrix<> result(M1.rows(), M2.cols());
   typename Rows<_Matrix1>::const_iterator r1=rows(M1).begin();
   for (typename Entire< Rows< IncidenceMatrix<> > >::iterator dst=entire(rows(result));
	!dst.at_end();  ++dst, ++r1)
      accumulate_in(entire(rows(M2.minor(*r1,All))), BuildBinary<operations::add>(), *dst);

   return result;
}

template <typename Permutation> inline
IncidenceMatrix<> permuted_rows(const IncidenceMatrix<>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_rows - dimension mismatch");
   RestrictedIncidenceMatrix<sparse2d::only_rows> result(M.rows(), select(rows(M),perm).begin());
   return IncidenceMatrix<>(result);
}

template <typename Permutation> inline
IncidenceMatrix<> permuted_cols(const IncidenceMatrix<>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_cols - dimension mismatch");
   RestrictedIncidenceMatrix<sparse2d::only_cols> result(M.cols(), select(cols(M),perm).begin());
   return IncidenceMatrix<>(result);
}

template <typename Permutation> inline
IncidenceMatrix<> permuted_inv_rows(const IncidenceMatrix<>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_inv_rows - dimension mismatch");
   RestrictedIncidenceMatrix<sparse2d::only_rows> result(M.rows());
   std::copy(entire(rows(M)), select(rows(result),perm).begin());
   return IncidenceMatrix<>(result);
}

template <typename Permutation> inline
IncidenceMatrix<> permuted_inv_cols(const IncidenceMatrix<>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_inv_cols - dimension mismatch");
   RestrictedIncidenceMatrix<sparse2d::only_cols> result(M.cols());
   std::copy(entire(cols(M)), select(cols(result),perm).begin());
   return IncidenceMatrix<>(result);
}

template <typename Permutation> inline
IncidenceMatrix<Symmetric> permuted_rows(const IncidenceMatrix<Symmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_rows - dimension mismatch");
   std::vector<int> inv_perm(M.rows());
   inverse_permutation(perm,inv_perm);
   return M.copy_permuted(perm,inv_perm);
}

template <typename Permutation> inline
IncidenceMatrix<Symmetric>
permuted_inv_rows(const IncidenceMatrix<Symmetric>& M, const Permutation& inv_perm,
		  typename enable_if<int, derived_from<typename container_traits<Permutation>::category, random_access_iterator_tag>::value>::type=0)
{
   DIMENSION_CHECK(M.rows() != inv_perm.size(), "permuted_inv_rows - dimension mismatch");
   std::vector<int> perm(M.rows());
   inverse_permutation(inv_perm,perm);
   return M.copy_permuted(perm,inv_perm);
}

template <typename Permutation> inline
IncidenceMatrix<Symmetric>
permuted_inv_rows(const IncidenceMatrix<Symmetric>& M, const Permutation& inv_perm,
		  typename disable_if<int, derived_from<typename container_traits<Permutation>::category, random_access_iterator_tag>::value>::type=0)
{
   DIMENSION_CHECK(M.rows() != inv_perm.size(), "permuted_inv_rows - dimension mismatch");
   std::vector<int> inv_perm_copy(inv_perm.size());
   std::copy(entire(inv_perm), inv_perm_copy.begin());
   return permuted_inv_rows(M,inv_perm_copy);
}

template <typename Permutation> inline
IncidenceMatrix<Symmetric> permuted_cols(const IncidenceMatrix<Symmetric>& M, const Permutation& perm)
{
   return permuted_rows(M,perm);
}

template <typename Permutation> inline
IncidenceMatrix<Symmetric> permuted_inv_cols(const IncidenceMatrix<Symmetric>& M, const Permutation& inv_perm)
{
   return permuted_inv_rows(M,inv_perm);
}

} // end namespace pm

namespace polymake {
   using pm::IncidenceMatrix;
   using pm::RestrictedIncidenceMatrix;
}

namespace std {
   template <typename symmetric> inline
   void swap(pm::IncidenceMatrix<symmetric>& M1, pm::IncidenceMatrix<symmetric>& M2) { M1.swap(M2); }

   template <pm::sparse2d::restriction_kind restriction> inline
   void swap(pm::RestrictedIncidenceMatrix<restriction>& M1,
	     pm::RestrictedIncidenceMatrix<restriction>& M2) { M1.swap(M2); }

   template <typename Tree> inline
   void swap(pm::incidence_line<Tree>& l1, pm::incidence_line<Tree>& l2) { l1.swap(l2); }
}

#endif // _POLYMAKE_INCIDENCE_MATRIX_H

// Local Variables:
// mode:C++
// End:
