/* Copyright (c) 1997-2009
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_MATRIX_H
#define _POLYMAKE_MATRIX_H "$Project: polymake $$Id: Matrix.h 9089 2009-02-19 17:37:14Z gawrilow $"

#ifndef _POLYMAKE_GENERIC_MATRIX_H
#  include <GenericMatrix.h>
#endif
#ifndef _POLYMAKE_SHARED_OBJECT_H
#  include <shared_object.h>
#endif

namespace pm {

template <typename Eref, bool rowwise> class matrix_line_factory;

template <typename E>
class Matrix_base {
protected:
   struct dim_t {
      int dimr, dimc;
      dim_t() : dimr(0), dimc(0) {}
      dim_t(int r, int c) : dimr(r), dimc(c) {}
   };
   typedef shared_array<E, list( PrefixData<dim_t>, AliasHandler<shared_alias_handler> ) > shared_array_type;
   shared_array_type data;

   friend Matrix_base& make_mutable_alias(Matrix_base& alias, Matrix_base& owner)
   {
      alias.data.make_mutable_alias(owner.data);
      return alias;
   }
   template <typename,bool> friend class matrix_line_factory;

   Matrix_base() {}

   Matrix_base(int r, int c) : data(dim_t(r,c), r*c) {}

   template <typename Iterator>
   Matrix_base(int r, int c, Iterator src) : data(dim_t(r,c), r*c, src) {}

   Matrix_base(void *place, int r, int c) : data(place, dim_t(r,c), r*c) {}

   template <typename Iterator>
   Matrix_base(void *place, int r, int c, Iterator src) : data(place, dim_t(r,c), r*c, src) {}
};

template <typename E>
class Matrix : public GenericMatrix<Matrix<E>, E>,
	       protected Matrix_base<E> {
   typedef Matrix_base<E> base;
   typedef typename base::dim_t dim_t;

   friend Matrix& make_mutable_alias(Matrix& alias, Matrix& owner)
   {
      return static_cast<Matrix&>(make_mutable_alias(static_cast<base&>(alias), static_cast<base&>(owner)));;
   }

   template <typename,bool> friend class matrix_line_factory;
public:
   typedef E value_type;
   typedef E& reference;
   typedef const E& const_reference;
   static const bool prefer_concat=true;

   Matrix() {}

   Matrix(int r, int c) : base(r,c) {}

   Matrix(int r, int c, typename function_argument<E>::type init)
      : base(r, c, constant(init).begin()) {}

   /** Matrix with given dimensions.  Elements are initialized from an input sequence.
       Elements are assumed to come in the row order.
   */
   template <typename Iterator>
   Matrix(int r, int c, Iterator src,
	  typename enable_if<nothing*, construct_cascade_helper<Iterator,E>::depth==1>::type=0)
      : base(r, c, src) {}

   template <size_t r, size_t c>
   explicit Matrix(const E (&a)[r][c])
      : base(r, c, &a[0][0]) {}

   template <typename Iterator>
   Matrix(int r, int c, Iterator src,
	  typename enable_if<int, (construct_cascade_helper<Iterator,E>::depth==2 &&
				   (check_iterator_feature<Iterator,end_sensitive>::value ||
				    check_iterator_feature<Iterator,unlimited>::value))>::type=0)
      : base(r, c, construct_matching_iterator<E>(src, (dense*)0)) {}

   /// Copy of a disguised Matrix object.
   Matrix(const GenericMatrix<Matrix>& m) : base(m.top()) {}

   /// Copy of an abstract matrix of the same element type.
   template <typename Matrix2>
   Matrix(const GenericMatrix<Matrix2,E>& m)
      : base(m.rows(), m.cols(), ensure(concat_rows(m), (dense*)0).begin()) {}

   /// Copy of an abstract matrix with element conversion.
   template <typename Matrix2, typename E2>
   explicit Matrix(const GenericMatrix<Matrix2,E2>& m)
      : base(m.rows(), m.cols(), ensure(attach_converter<E>(concat_rows(m)), (dense*)0).begin()) {}

   template <typename Container>
   Matrix(const Container& src,
	  typename enable_if<nothing*, isomorphic_to_container_of<Container, Vector<E> >::value>::type=0)
      : base(src.size(), src.empty() ? 0 : get_dim(src.front()), make_cascade_iterator(src, (dense*)0, int2type<2>())) {}

protected:
   Matrix(void *place, int r, int c) : base(place,r,c) {}

   template <typename Iterator>
   Matrix(void *place, int r, int c, Iterator src,
	  typename enable_if<nothing*, construct_cascade_helper<Iterator,E>::depth==1>::type=0)
      : base(place,r,c,src) {}

   template <typename Iterator>
   Matrix(void *place, int r, int c, Iterator src,
	  typename enable_if<int, (construct_cascade_helper<Iterator,E>::depth==2 &&
				   (check_iterator_feature<Iterator,end_sensitive>::value ||
				    check_iterator_feature<Iterator,unlimited>::value))>::type=0)
      : base(place,r,c,construct_matching_iterator<E>(src, (dense*)0)) {}

   template <typename Container>
   Matrix(void *place, int r, int c, const Container& src,
	  typename enable_if<nothing*, isomorphic_to_container_of<Container, Vector<E> >::value>::type=0)
      : base(place,r,c,make_cascade_iterator(src, (dense*)0, int2type<2>())) {}

   void resize(void *place, int r, int c)
   {
      this->data.resize(place,r*c);
      this->data.get_prefix().dimr=r;
      this->data.get_prefix().dimc=c;
   }
public:
   Matrix& operator= (const Matrix& m) { this->data=m.data; return *this; }

   using Matrix::generic_type::operator=;

   void swap(Matrix& m) { this->data.swap(m.data); }

   friend void relocate(Matrix* from, Matrix* to)
   {
      relocate(&from->data, &to->data);
   }

   /// Resize to new dimensions, added elements initialized with default constructor.
   void resize(int r, int c)
   {
      const int dimc=cols(), dimr=rows();
      if (c==dimc) {
	 this->data.resize(r*c);
	 this->data.get_prefix().dimr=r;
      } else if (c<dimc && r<=dimr) {
	 *this=this->minor(sequence(0,r),sequence(0,c));
      } else {
	 Matrix M(r,c);
	 if (c<dimc)
	    M.minor(sequence(0,dimr),All)=this->minor(All,sequence(0,c));
	 else
	    M.minor(sequence(0,std::min(dimr,r)), sequence(0,dimc))=this->minor(sequence(0,std::min(dimr,r)),All);
	 *this=M;
      }
   }

   template <typename E2>
   void assign(int r, int c, const E2& x)
   {
      this->data.assign(r*c, constant(x).begin());
      this->data.get_prefix()=dim_t(r,c);
   }

   void clear() { this->data.clear(); }
   void clear(int r, int c)
   {
      this->data.resize(r*c);
      this->data.get_prefix()=dim_t(r,c);
   }

   int rows() const { return this->data.get_prefix().dimr; }
   int cols() const { return this->data.get_prefix().dimc; }

   E& operator() (int i, int j)
   {
      INDEX_CHECK((i<0 || i>=rows()), "operator() - row index out of range");
      INDEX_CHECK((j<0 || j>=cols()), "operator() - column index out of range");
      return (*this->data)[i*cols()+j];
   }

   const E& operator() (int i, int j) const
   {
      INDEX_CHECK((i<0 || i>=rows()), "operator() - row index out of range");
      INDEX_CHECK((j<0 || j>=cols()), "operator() - column index out of range");
      return (*this->data)[i*cols()+j];
   }

   template <typename Matrix2>
   Matrix& operator/= (const GenericMatrix<Matrix2,E>& m)
   {
      const int add_rows=m.rows(), add_cols=m.cols();
      if (add_rows) {
	 DIMENSION_CHECK((this->data.size() && cols() != add_cols), "operator/= - column dimensions mismatch");
	 this->data.append(add_rows*add_cols, ensure(concat_rows(m), (dense*)0).begin());
	 this->data.get_prefix().dimr+=add_rows;
	 this->data.get_prefix().dimc=add_cols;
      }
      return *this;
   }

   template <typename Vector2>
   Matrix& operator/= (const GenericVector<Vector2,E>& v)
   {
      const int add_cols=v.dim();
      DIMENSION_CHECK((this->data.size() && cols() != add_cols), "operator/= - column dimensions mismatch");
      this->data.append(add_cols, ensure(v.top(), (dense*)0).begin());
      this->data.get_prefix().dimr++;
      this->data.get_prefix().dimc=add_cols;
      return *this;
   }

   using Matrix::generic_type::operator/=;

   template <typename Matrix2>
   Matrix& operator|= (const GenericMatrix<Matrix2,E>& m)
   {
      const int add_rows=m.rows(), add_cols=m.cols();
      if (add_cols) {
	 DIMENSION_CHECK((this->data.size() && rows() != add_rows), "operator|= - row dimensions mismatch");
	 this->data.weave(add_rows*add_cols, cols(), pm::rows(m).begin());
	 this->data.get_prefix().dimr=add_rows;
	 this->data.get_prefix().dimc+=add_cols;
      }
      return *this;
   }

   template <typename Vector2>
   Matrix& operator|= (const GenericVector<Vector2,E>& v)
   {
      const int add_rows=v.dim();
      DIMENSION_CHECK((this->data.size() && rows() != add_rows), "operator|= - row dimensions mismatch");
      this->data.weave(add_rows, cols(), ensure(v.top(), (dense*)0).begin());
      this->data.get_prefix().dimr=add_rows;
      this->data.get_prefix().dimc++;
      return *this;
   }

protected:
   friend class ConcatRows<Matrix>;
   friend class GenericMatrix<Matrix>;
   friend class Rows<Matrix>;
   friend class Cols<Matrix>;

   void assign(const GenericMatrix<Matrix>& m) { this->data=m.top().data; }

   template <typename Matrix2>
   void assign(const GenericMatrix<Matrix2>& m)
   {
      const int r=m.rows(), c=m.cols();
      this->data.assign(r*c, ensure(concat_rows(m), (dense*)0).begin());
      this->data.get_prefix().dimr=r;
      this->data.get_prefix().dimc=c;
   }

   template <typename Operation>
   void assign_op(const Operation& op)
   {
      this->data.assign_op(op);
   }

   template <typename Source2, typename Operation>
   void assign_op(const Source2& src2, const Operation& op)
   {
      this->data.assign_op(ensure(concat_rows(src2), (dense*)0).begin(), op);
   }

   template <typename E2>
   void _fill(const E2& x, True)
   {
      this->data.assign(this->data.size(), constant(x).begin());
   }
};

template <typename E>
class ConcatRows< Matrix<E> >
   : public plain_array< ConcatRows< Matrix<E> >, E >,
     public GenericVector< ConcatRows< Matrix<E> >, E> {
protected:
   ~ConcatRows();
   Matrix<E>& hidden() { return *reinterpret_cast<Matrix<E>*>(this); }
   const Matrix<E>& hidden() const { return *reinterpret_cast<const Matrix<E>*>(this); }
public:
   int size() const { return hidden().data.size(); }
   E& front() { return **hidden().data; }
   const E& front() const { return **hidden().data; }

   ConcatRows& operator= (const ConcatRows& m)
   {
      return ConcatRows::generic_type::operator=(m);
   }
   using ConcatRows::generic_type::operator=;
};

template <typename Eref, bool rowwise>
class matrix_line_factory {
protected:
   typedef typename deref<Eref>::type E;
   typedef typename inherit_ref<Matrix_base<E>, Eref>::type matrix_base_ref;
   typedef typename inherit_ref<Matrix<E>, Eref>::type matrix_ref;

   mutable alias<matrix_base_ref> matrix;

   template <typename> friend class Rows;
   template <typename> friend class Cols;
   template <typename,bool> friend class matrix_line_factory;

   matrix_line_factory(matrix_base_ref m_arg) : matrix(m_arg) {}
public:
   typedef int argument_type;
   typedef IndexedSlice<masquerade<ConcatRows, matrix_ref>, Series<int,rowwise> > result_type;

   matrix_line_factory() : matrix(Matrix_base<E>()) {}
   matrix_line_factory(const matrix_line_factory<typename attrib<Eref>::minus_const, rowwise>& f) : matrix(f.matrix) {}

   result_type operator() (int start) const
   {
      matrix_ref me=static_cast<matrix_ref>(static_cast<matrix_base_ref>(matrix));
      return result_type(me, Series<int,rowwise>(start, rowwise ? me.cols() : me.rows(), rowwise ? 1 : me.cols()));
   }
};

template <typename Eref, bool rowwise>
struct operation_cross_const_helper< matrix_line_factory<Eref,rowwise> > {
   typedef matrix_line_factory<typename attrib<Eref>::minus_const, rowwise> operation;
   typedef matrix_line_factory<typename attrib<Eref>::plus_const, rowwise> const_operation;
};

template <typename E>
class Rows< Matrix<E> >
   : public modified_container_impl< Rows< Matrix<E> >,
				     list( Container< series >,
					   Operation< matrix_line_factory<E&,true> >,
					   MasqueradedTop ) > {
protected:
   ~Rows();
public:
   series get_container() const
   {
      const Matrix<E>& me=this->hidden();
      return series(0, me.rows(), me.cols());
   }
   matrix_line_factory<E&,true> get_operation() { return this->hidden(); }
   matrix_line_factory<const E&,true> get_operation() const { return this->hidden(); }

   void resize(int n)
   {
      Matrix<E>& me=this->hidden();
      me.resize(n, me.cols());
   }
};

template <typename E>
class Cols< Matrix<E> >
   : public modified_container_impl< Cols< Matrix<E> >,
				     list( Container< sequence >,
					   Operation< matrix_line_factory<E&,false> >,
					   MasqueradedTop ) > {
protected:
   ~Cols();
public:
   sequence get_container() const
   {
      return sequence(0, this->hidden().cols());
   }
   matrix_line_factory<E&,false> get_operation() { return this->hidden(); }
   matrix_line_factory<const E&,false> get_operation() const { return this->hidden(); }

   void resize(int n)
   {
      Matrix<E>& me=this->hidden();
      me.resize(me.rows(), n);
   }
};

template <typename E, typename Permutation> inline
Matrix<E> permuted_rows(const Matrix<E>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_rows - dimension mismatch");
   return Matrix<E>(M.rows(), M.cols(), select(rows(M),perm).begin());
}

template <typename E, typename Permutation> inline
Matrix<E> permuted_cols(const Matrix<E>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_cols - dimension mismatch");
   Matrix<E> result(M.rows(), M.cols());
   std::copy(entire(select(cols(M),perm)), cols(result).begin());
   return result;
}

template <typename E, typename Permutation> inline
Matrix<E> permuted_inv_rows(const Matrix<E>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_inv_rows - dimension mismatch");
   Matrix<E> result(M.rows(), M.cols());
   std::copy(entire(rows(M)), select(rows(result), perm).begin());
   return result;
}

template <typename E, typename Permutation> inline
Matrix<E> permuted_inv_cols(const Matrix<E>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_inv_cols - dimension mismatch");
   Matrix<E> result(M.rows(), M.cols());
   std::copy(entire(cols(M)), select(cols(result), perm).begin());
   return result;
}

} // end namespace pm

namespace polymake {
   using pm::Matrix;
}

namespace std {
   template <typename E> inline
   void swap(pm::Matrix<E>& m1, pm::Matrix<E>& m2) { m1.swap(m2); }
}

#endif // _POLYMAKE_MATRIX_H

// Local Variables:
// mode:C++
// End:
