/* Copyright (c) 1997-2009
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_SPARSE_MATRIX_H
#define _POLYMAKE_SPARSE_MATRIX_H "$Project: polymake $$Id: SparseMatrix.h 9022 2009-01-05 23:35:32Z gawrilow $"

#ifndef _POLYMAKE_SPARSE2D_H
#  include <sparse2d.h>
#endif
#ifndef _POLYMAKE_MATRIX_H
#  include <Matrix.h>
#endif
#ifndef _POLYMAKE_SPARSE_VECTOR_H
#  include <SparseVector.h>
#endif

namespace pm {

template <typename Tree, typename symmetric>
class sparse_matrix_line
   : public sparse2d::line<Tree>,
     public GenericVector< sparse_matrix_line<Tree,symmetric>, typename Tree::mapped_type> {
protected:
   ~sparse_matrix_line();
public:
   typedef typename Tree::mapped_type value_type;
   friend class SparseMatrix<value_type,symmetric>;
   friend class GenericVector<sparse_matrix_line>;
   template <typename,typename> friend class GenericMatrix;
   template <typename,typename> friend class sparse_matrix_line;

   sparse_matrix_line& operator= (sparse_matrix_line& l)
   {
      return sparse_matrix_line::generic_type::operator=(l);
   }

   using sparse_matrix_line::generic_type::operator=;
protected:
   typedef sparse_proxy_base< sparse2d::line<Tree> > proxy_base;

   template <typename Container>
   void assign(const Container& c, Symmetric)
   {
      assign_sparse(sparse2d::select_lower_triangle(*this),
		    attach_truncator(ensure(c, (pure_sparse*)0), index_truncator(this->get_line_index())));
   }

   template <typename Operation>
   void assign_op(const Operation& op, Symmetric)
   {
      perform_assign(entire(sparse2d::select_lower_triangle(*this)), op);
   }

   template <typename Container, typename Operation>
   typename disable_if<void, operations::is_partially_defined_for<Operation,sparse_matrix_line,Container>::value>::type
   assign_op(const Container& c, const Operation& op, Symmetric)
   {
      perform_assign(entire(sparse2d::select_lower_triangle(*this)), c.begin(), op);
   }

   template <typename Container, typename Operation>
   typename enable_if<void, operations::is_partially_defined_for<Operation,sparse_matrix_line,Container>::value>::type
   assign_op(const Container& c, const Operation& op, Symmetric)
   {
      perform_assign_sparse(sparse2d::select_lower_triangle(*this),
			    attach_truncator(ensure(c, (pure_sparse*)0), index_truncator(this->get_line_index())), op);
   }

   using sparse_matrix_line::generic_type::assign;
   using sparse_matrix_line::generic_type::assign_op;
public:

   typedef sparse_elem_proxy<proxy_base> reference;
   typedef const value_type& const_reference;
   typedef random_access_iterator_tag container_category;

   reference operator[] (int i) { return proxy_base(*this,i); }

   const_reference operator[] (int i) const
   {
      operations::clear<value_type> zero;
      return zero.deref_sparse(this->find(i));
   }

   int dim() const { return this->max_size(); }
};

template <typename Iterator=void>
class skew_negator {
protected:
   int diag;
   operations::neg<typename iterator_traits<Iterator>::reference> op;
public:
   skew_negator(int diag_arg=-1) : diag(diag_arg) {}

   typedef Iterator argument_type;
   typedef typename iterator_traits<Iterator>::value_type result_type;

   result_type operator() (const argument_type& it) const
   {
      if (it.index() > diag) return op(*it);
      return *it;
   }
};

template <>
class skew_negator<void> : operations::incomplete {
protected:
   int diag;
public:
   skew_negator(int diag_arg=-1) : diag(diag_arg) {}
   operator int () const { return diag; }
};

template <typename Iterator, typename Reference>
struct unary_op_builder< skew_negator<void>, Iterator, Reference > {
   typedef skew_negator<Iterator> operation;
   static operation create(int diag_arg) { return operation(diag_arg); }
};

template <typename Tree>
class sparse_matrix_line<Tree, SkewSymmetric>
   : public modified_container_impl< sparse_matrix_line<Tree,SkewSymmetric>,
				     list( Hidden< sparse_matrix_line<Tree,Symmetric> >,
					   Operation< skew_negator<> > ) >,
     public GenericVector< sparse_matrix_line<Tree,SkewSymmetric>, typename Tree::mapped_type> {
   typedef modified_container_impl<sparse_matrix_line> _super;
protected:
   ~sparse_matrix_line();
public:
   typedef typename Tree::mapped_type value_type;
   friend class SparseMatrix<value_type,SkewSymmetric>;
   friend class GenericVector<sparse_matrix_line>;
   template <typename,typename> friend class GenericMatrix;

   typename _super::operation get_operation() const { return get_line_index(); }

   sparse_matrix_line& operator= (sparse_matrix_line& l)
   {
      return GenericVector<sparse_matrix_line>::operator=(l);
   }

   using sparse_matrix_line::generic_type::operator=;
protected:
   typedef sparse_proxy_base< sparse2d::line<Tree> > proxy_base;

   void push_back(int i, typename function_argument<value_type>::type x) { this->hidden().push_back(i,x); }

   void _fill(typename function_argument<value_type>::type x, pure_sparse)
   {
      if (x)
	 fill_sparse(this->hidden(), attach_operation(ensure(constant(x), (indexed*)0), get_operation()).begin());
      else
	 clear();
   }

   template <typename Container>
   void assign(const Container& c)
   {
      assign_sparse(this->hidden(), attach_operation(ensure(c, (pure_sparse*)0), get_operation()).begin());
   }
   template <typename Operation>
   void assign_op(const Operation& op)
   {
      this->hidden().assign_op(op);
   }
   template <typename Container, typename Operation>
   typename disable_if<void, operations::is_partially_defined_for<Operation,sparse_matrix_line,Container>::value>::type
   assign_op(const Container& c, const Operation& op)
   {
      this->hidden().assign_op(c,op);
   }
   template <typename Container, typename Operation>
   typename enable_if<void, operations::is_partially_defined_for<Operation,sparse_matrix_line,Container>::value>::type
   assign_op(const Container& c, const Operation& op)
   {
      this->hidden().assign_op(attach_operation(ensure(c, (pure_sparse*)0), get_operation()), op);
   }

   template <typename Container>
   void assign(const Container& c, SkewSymmetric)
   {
      this->hidden().assign(c, Symmetric());
   }
   template <typename Operation>
   void assign_op(const Operation& op, SkewSymmetric)
   {
      this->hidden().assign_op(op, Symmetric());
   }
   template <typename Container, typename Operation>
   void assign_op(const Container& c, const Operation& op, SkewSymmetric)
   {
      this->hidden().assign_op(c, op, Symmetric());
   }

public:
   typedef sparse_elem_proxy<proxy_base, value_type, SkewSymmetric> reference;
   typedef const value_type const_reference;
   typedef random_access_iterator_tag container_category;

   reference operator[] (int i) { return proxy_base(this->hidden(),i); }

   const_reference operator[] (int i) const
   {
      operations::clear<value_type> zero;
      return zero.deref_sparse(find(i));
   }

   typename _super::iterator find(int i)
   {
      return typename _super::iterator(this->hidden().find(i), get_operation());
   }
   typename _super::const_iterator find(int i) const
   {
      return typename _super::const_iterator(this->hidden().find(i), get_operation());
   }
   typename _super::iterator insert(int i, const value_type& x)
   {
      operations::neg<const value_type&> op;
      if (i>get_line_index())
	 return typename _super::iterator(this->hidden().insert(i,op(x)), get_operation());
      return typename _super::iterator(this->hidden().insert(i,x));
   }

   typename _super::iterator insert(const typename _super::iterator& pos, int i, const value_type& x)
   {
      operations::neg<const value_type&> op;
      if (i>get_line_index())
	 return typename _super::iterator(this->hidden().insert(pos,i,op(x)), get_operation());
      return typename _super::iterator(this->hidden().insert(pos,i,x), get_operation());
   }

   void erase(int i) { this->hidden().erase(i); }
   void erase(const typename _super::iterator& pos) { this->hidden().erase(pos); }

   void clear() { this->hidden().clear(); }
   int dim() const { return this->hidden().dim(); }
   int get_line_index() const { return this->hidden().get_line_index(); }
};

template <typename tree_type, typename symmetric>
struct check_container_feature<sparse_matrix_line<tree_type,symmetric>, pure_sparse> : True {};

template <typename tree_type, typename symmetric>
struct spec_object_traits< sparse_matrix_line<tree_type,symmetric> >
   : spec_object_traits<is_container> {
   typedef tree_type masquerade_for;
   static const int is_resizeable=0;
};


template <typename E, sparse2d::restriction_kind restriction=sparse2d::only_rows>
class RestrictedSparseMatrix;

template <typename E, sparse2d::restriction_kind restriction>
class RestrictedSparseMatrix
   : public matrix_methods<RestrictedSparseMatrix<E,restriction>, E> {
protected:
   typedef sparse2d::Table<E, false, restriction> table_type;
   table_type data;

   template <typename Iterator>
   void _copy(Iterator src, True) { std::copy(src, entire(pm::rows(*this))); }

   template <typename Iterator>
   void _copy(Iterator src, False) { std::copy(src, entire(pm::cols(*this))); }

   typedef sparse_matrix_line<typename table_type::primary_tree_type, NonSymmetric> _line;
public:
   typedef E value_type;
   typedef typename _line::reference reference;
   typedef const E& const_reference;

   explicit RestrictedSparseMatrix(int n=0) : data(n) {}

   template <typename Iterator>
   RestrictedSparseMatrix(int n, Iterator src)
      : data(n)
   {
      _copy(src, bool2type<restriction==sparse2d::only_rows>());
   }

   void swap(RestrictedSparseMatrix& M)
   {
      data.swap(M.data);
   }

   void clear() { data.clear(); }

protected:
   reference _random(int i, int j, False)
   {
      return this->row(i)[j];
   }
   reference _random(int i, int j, True)
   {
      return this->col(j)[i];
   }
   const_reference _random(int i, int j, False) const
   {
      return this->row(i)[j];
   }
   const_reference _random(int i, int j, True) const
   {
      return this->col(j)[i];
   }
public:
   reference operator() (int i, int j)
   {
      return _random(i, j, bool2type<restriction==sparse2d::only_cols>());
   }
   const_reference operator() (int i, int j) const
   {
      return _random(i, j, bool2type<restriction==sparse2d::only_cols>());
   }

private:
   template <typename RowCol, typename _Vector>
   void _append(RowCol& row_col, const _Vector& vec, int i)
   {
      for (typename Entire<_Vector>::const_iterator v=entire(vec); !v.at_end(); ++v)
	 row_col[v.index()].push_back(i,*v);
   }

   template <typename Iterator>
   void _append_rows(int n, Iterator src, True)
   {
      int oldrows=data.rows();
      data.resize_rows(oldrows+n);
      std::copy(src, pm::rows(*this).begin()+oldrows);
   }

   template <typename Iterator>
   void _append_rows(int, Iterator src, False)
   {
      for (int r=data.rows(); !src.at_end(); ++src, ++r)
	 _append(pm::cols(*this), *src, r);
   }

   template <typename Iterator>
   void _append_cols(int n, Iterator src, True)
   {
      int oldcols=data.cols();
      data.resize_cols(oldcols+n);
      std::copy(src, pm::cols(*this).begin()+oldcols);
   }

   template <typename Iterator>
   void _append_cols(int, Iterator src, False)
   {
      for (int c=data.cols(); !src.at_end(); ++src, ++c)
	 _append(pm::rows(*this), *src, c);
   }

public:
   template <typename _Matrix>
   RestrictedSparseMatrix& operator/= (const GenericMatrix<_Matrix>& m)
   {
      _append_rows(m.rows(), entire(pm::rows(m)), bool2type<restriction==sparse2d::only_rows>());
      return *this;
   }

   template <typename _Vector>
   RestrictedSparseMatrix& operator/= (const GenericVector<_Vector>& v)
   {
      _append_rows(1, entire(item2container(v.top())), bool2type<restriction==sparse2d::only_rows>());
      return *this;
   }

   template <typename _Matrix>
   RestrictedSparseMatrix& operator|= (const GenericMatrix<_Matrix>& m)
   {
      _append_cols(m.cols(), entire(pm::cols(m)), bool2type<restriction==sparse2d::only_cols>());
      return *this;
   }

   template <typename _Vector>
   RestrictedSparseMatrix& operator|= (const GenericVector<_Vector>& v)
   {
      _append_cols(1, entire(item2container(v.top())), bool2type<restriction==sparse2d::only_cols>());
      return *this;
   }

   void squeeze() { data.squeeze(); }

   template <typename Iterator>
   void permute_rows(Iterator perm)
   {
      data.permute_rows(perm, False());
   }

   template <typename Iterator>
   void permute_cols(Iterator perm)
   {
      data.permute_cols(perm, False());
   }

   template <typename Iterator>
   void permute_inv_rows(Iterator inv_perm)
   {
      data.permute_rows(inv_perm, True());
   }

   template <typename Iterator>
   void permute_inv_cols(Iterator inv_perm)
   {
      data.permute_cols(inv_perm, True());
   }

#ifdef POLY_DEBUG
   void check() const { data.check(); }
#endif
   friend class Rows<RestrictedSparseMatrix>;
   friend class Cols<RestrictedSparseMatrix>;
   friend class SparseMatrix<E>;
};

template <typename E, sparse2d::restriction_kind restriction>
struct spec_object_traits< RestrictedSparseMatrix<E, restriction> >
   : spec_object_traits<is_container> {
   static const int dimension=2;
};

template <typename E, sparse2d::restriction_kind restriction>
class Rows< RestrictedSparseMatrix<E, restriction> >
   : public redirected_container< Rows< RestrictedSparseMatrix<E, restriction> >,
				  list( Container< sparse2d::Rows<E, false, restriction, operations::masquerade2<sparse_matrix_line,NonSymmetric> > >,
					MasqueradedTop ) > {
public:
   typedef typename if_else<restriction==sparse2d::only_rows, random_access_iterator_tag, output_iterator_tag>::type
      container_category;
   typename Rows::container& get_container()
   {
      return reinterpret_cast<typename Rows::container&>(this->hidden().data);
   }
   const typename Rows::container& get_container() const
   {
      return reinterpret_cast<const typename Rows::container&>(this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data.resize_rows(n);
   }
};

template <typename E, sparse2d::restriction_kind restriction>
class Cols< RestrictedSparseMatrix<E, restriction> >
   : public redirected_container< Cols< RestrictedSparseMatrix<E, restriction> >,
				  list( Container< sparse2d::Cols<E, false, restriction, operations::masquerade2<sparse_matrix_line,NonSymmetric> > >,
					MasqueradedTop ) > {
public:
   typedef typename if_else<restriction==sparse2d::only_cols, random_access_iterator_tag, output_iterator_tag>::type
      container_category;
   typename Cols::container& get_container()
   {
      return reinterpret_cast<typename Cols::container&>(this->hidden().data);
   }
   const typename Cols::container& get_container() const
   {
      return reinterpret_cast<const typename Cols::container&>(this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data.resize_cols(n);
   }
};

template <typename E, typename symmetric>
class SparseMatrix : public GenericMatrix< SparseMatrix<E,symmetric>, E> {
protected:
   typedef sparse2d::Table<E, symmetric::value> table_type;
   shared_object<table_type, AliasHandler<shared_alias_handler> > data;

   friend SparseMatrix& make_mutable_alias(SparseMatrix& alias, SparseMatrix& owner)
   {
      alias.data.make_mutable_alias(owner.data);
      return alias;
   }

   // elementwise, non-symmetric
   template <typename Iterator>
   void _init(Iterator src, True, False)
   {
      conv<E,bool> non_zero;
      const int n=this->cols();
      for (typename Entire< Rows<SparseMatrix> >::iterator r_i=entire(pm::rows(*this)); !r_i.at_end(); ++r_i)
	 for (int i=0; i<n; ++i, ++src)
	    if (non_zero(*src)) r_i->push_back(i,*src);
   }

   // elementwise, symmetric
   template <typename Iterator>
   void _init(Iterator src, True, True)
   {
      conv<E,bool> non_zero;
      const int n=this->cols();
      int d=0;
      for (typename Entire< Rows<SparseMatrix> >::iterator r_i=entire(pm::rows(*this)); !r_i.at_end(); ++r_i) {
	 for (int i=0; i<=d; ++i, ++src)
	    if (non_zero(*src)) r_i->push_back(i,*src);
	 ++d; std::advance(src,n-d);
      }
   }

   // rowwise, non-symmetric
   template <typename Iterator>
   void _init(Iterator src, False, False)
   {
      std::copy(src, entire(pm::rows(*this)));
   }

   template <typename Row, typename Iterator>
   void _init_row(Row& r, Iterator src, int d)
   {
      for (int i; !src.at_end() && (i=src.index())<=d; ++src)
	 r.push_back(i,*src);
   }

   // rowwise, symmetric
   template <typename Iterator>
   void _init(Iterator src, False, True)
   {
      int d=0;
      for (typename Entire< Rows<SparseMatrix> >::iterator r_i=entire(pm::rows(*this)); !r_i.at_end(); ++r_i, ++d, ++src)
	 _init_row(*r_i, ensure(*src, (pure_sparse*)0).begin(), d);
   }

   typedef sparse_matrix_line<typename table_type::primary_tree_type, symmetric> _line;
public:
   typedef E value_type;
   typedef typename _line::reference reference;
   typedef typename _line::const_reference const_reference;

   SparseMatrix() {}

   SparseMatrix(int m, int n)
      : data( make_constructor(m,n,(table_type*)0) ) {}

   template <typename Iterator>
   SparseMatrix(int m, int n, Iterator src)
      : data( make_constructor(m,n,(table_type*)0) )
   {
      _init(src, bool2type<(object_traits<typename iterator_traits<Iterator>::value_type>::total_dimension ==
			    object_traits<E>::total_dimension)>(), symmetric());
   }

   SparseMatrix(const GenericMatrix<SparseMatrix>& M)
      : data(M.top().data) {}

   template <typename Matrix2>
   SparseMatrix(const GenericMatrix<Matrix2,E>& M,
		typename enable_if<nothing*, !symmetric::value || identical<symmetric, typename matrix_symmetry_type<Matrix2>::type>::value>::type=0)
      : data( make_constructor(M.rows(), M.cols(), (table_type*)0) )
   {
      _init(pm::rows(M).begin(), False(), symmetric());
   }

   template <typename Matrix2, typename E2>
   explicit SparseMatrix(const GenericMatrix<Matrix2,E2>& M,
			 typename enable_if<nothing*, !symmetric::value || identical<symmetric, typename matrix_symmetry_type<Matrix2>::type>::value>::type=0)
      : data( make_constructor(M.rows(), M.cols(), (table_type*)0) )
   {
      _init(pm::rows(M).begin(), False(), symmetric());
   }

   template <sparse2d::restriction_kind restriction>
   explicit SparseMatrix(RestrictedSparseMatrix<E,restriction>& M,
			 typename disable_if<typename cons<nothing*, RestrictedSparseMatrix<E,restriction> >::head, symmetric::value>::type=0)
      : data( constructor<table_type(typename RestrictedSparseMatrix<E,restriction>::table_type&)>(M.data) ) {}

   template <typename Container>
   SparseMatrix(const Container& src,
		typename enable_if<nothing*, isomorphic_to_container_of<Container, Vector<E> >::value &&
			                     !symmetric::value>::type=0)
      : data( make_constructor(src.size(), src.empty() ? 0 : get_dim(src.front()), (table_type*)0) )
   {
      _init(src.begin(), False(), symmetric());
   }

   SparseMatrix& operator= (const SparseMatrix& M) { data=M.data; return *this; }

   using SparseMatrix::generic_type::operator=;

   template <sparse2d::restriction_kind restriction>
   typename disable_if<typename cons<SparseMatrix, RestrictedSparseMatrix<E,restriction> >::head, symmetric::value>::type&
   operator= (RestrictedSparseMatrix<E,restriction>& M)
   {
      data=make_constructor(M.data, (table_type*)0);
      return *this;
   }

   void swap(SparseMatrix& M) { data.swap(M.data); }

   friend void relocate(SparseMatrix* from, SparseMatrix* to)
   {
      relocate(&from->data, &to->data);
   }

   void resize(int r, int c) { data->resize(r,c); }

   void clear() { data.apply(shared_clear()); }

   void clear(int r, int c) { data.apply(typename table_type::shared_clear(r,c)); }

   void remove0s()
   {
      for (typename Entire< Rows<SparseMatrix> >::iterator r=entire(pm::rows(*this)); !r.at_end(); ++r)
	 r->remove0s();
   }

   reference operator() (int i, int j) { return this->row(i)[j]; }

   const_reference operator() (int i, int j) const { return this->row(i)[j]; }

   template <typename Matrix2>
   typename disable_if<typename cons<SparseMatrix, Matrix2>::head, symmetric::value>::type&
   operator/= (const GenericMatrix<Matrix2,E>& M)
   {
      const int r=M.rows();
      if (r) {
	 DIMENSION_CHECK((this->cols() != M.cols()), "append_rows - column dimensions mismatch");
	 const int old_rows=this->rows();
	 data.apply(typename table_type::shared_add_rows(r));
	 copy(entire(pm::rows(M)), pm::rows(*this)+old_rows);
      }
      return *this;
   }

   template <typename Vector2>
   typename disable_if<typename cons<SparseMatrix, Vector2>::head, symmetric::value>::type&
   operator/= (const GenericVector<Vector2,E>& v)
   {
      DIMENSION_CHECK((this->cols() != v.dim()), "append_row - column dimensions mismatch");
      const int old_rows=this->rows();
      data.apply(typename table_type::shared_add_rows(1));
      this->row(old_rows)=v;
      return *this;
   }

   using SparseMatrix::generic_type::operator/=;

   template <typename Matrix2>
   typename disable_if<typename cons<SparseMatrix, Matrix2>::head, symmetric::value>::type&
   operator|= (const GenericMatrix<Matrix2,E>& M)
   {
      const int c=M.cols();
      if (c) {
	 DIMENSION_CHECK((this->rows() != M.rows()), "append_cols - row dimensions mismatch");
	 const int old_cols=this->cols();
	 data.apply(typename table_type::shared_add_cols(c));
	 copy(entire(pm::cols(M)), pm::cols(*this)+old_cols);
      }
      return *this;
   }

   template <typename Vector2>
   typename disable_if<typename cons<SparseMatrix, Vector2>::head, symmetric::value>::type&
   operator|= (const GenericVector<Vector2,E>& v)
   {
      DIMENSION_CHECK((this->rows() != v.dim()), "append_col - row dimensions mismatch");
      const int old_cols=this->cols();
      data.apply(typename table_type::shared_add_cols(1));
      this->col(old_cols)=v;
      return *this;
   }

   template <typename row_number_consumer, typename col_number_consumer>
   void squeeze(row_number_consumer rnc, col_number_consumer cnc) { data->squeeze(rnc,cnc); }

   template <typename row_number_consumer>
   void squeeze(row_number_consumer rnc) { data->squeeze(rnc); }

   void squeeze() { data->squeeze(); }

   template <typename row_number_consumer>
   void squeeze_rows(row_number_consumer rnc) { data->squeeze_rows(rnc); }

   void squeeze_rows() { data->squeeze_rows(); }

   template <typename col_number_consumer>
   void squeeze_cols(col_number_consumer cnc) { data->squeeze_cols(cnc); }

   void squeeze_cols() { data->squeeze_cols(); }

   template <typename Iterator>
   void permute_rows(Iterator perm)
   {
      data->permute_rows(perm, False());
   }

   template <typename Iterator>
   void permute_cols(Iterator perm)
   {
      data->permute_cols(perm, False());
   }

   template <typename Iterator>
   void permute_inv_rows(Iterator inv_perm)
   {
      data->permute_rows(inv_perm, True());
   }

   template <typename Iterator>
   void permute_inv_cols(Iterator inv_perm)
   {
      data->permute_cols(inv_perm, True());
   }

   template <typename Perm, typename InvPerm>
   SparseMatrix copy_permuted(const Perm& perm, const InvPerm& inv_perm,
			      typename enable_if<typename cons<nothing*,Perm>::head, symmetric::value>::type=0) const
   {
      const int n=this->rows();
      SparseMatrix result(n,n);
      result.data.get()->copy_permuted(*data, perm, inv_perm);
      return result;
   }

#ifdef POLY_DEBUG
   void check() const { data->check(); }
#endif

protected:
   friend class Rows<SparseMatrix>;
   friend class Cols<SparseMatrix>;
   friend class GenericMatrix<SparseMatrix>;

   void assign(const GenericMatrix<SparseMatrix>& m) { data=m.top().data; }

   template <typename Matrix2>
   void assign(const GenericMatrix<Matrix2>& m)
   {
      if (data.is_shared() || this->rows() != m.rows() || this->cols() != m.cols())
	 *this=SparseMatrix(m);
      else
	 SparseMatrix::generic_type::assign(m);
   }

   template <typename Operation>
   void assign_op(const Operation& op)
   {
      if (data.is_shared())
	 *this=SparseMatrix(LazyMatrix1<const SparseMatrix&, Operation>(*this,op));
      else
	 SparseMatrix::generic_type::assign_op(op);
   }

   template <typename Matrix2, typename Operation>
   void assign_op(const Matrix2& m, const Operation& op)
   {
      if (data.is_shared())
	 *this=SparseMatrix(LazyMatrix2<const SparseMatrix&, const Matrix2&, Operation>(*this,m,op));
      else
	 SparseMatrix::generic_type::assign_op(m,op);
   }

   template <typename E2>
   void _fill(const E2& x, False)
   {
      if (data.is_shared())
	 clear(this->rows(), this->cols());
      if (x)
	 SparseMatrix::generic_type::_fill(x, False());
   }
};

template <typename E, typename symmetric>
struct check_container_feature< SparseMatrix<E,symmetric>, pure_sparse > : True {};

template <typename E, typename symmetric>
struct check_container_feature< SparseMatrix<E,symmetric>, Symmetric > : identical<symmetric,Symmetric> {};

template <typename E, typename symmetric>
struct check_container_feature< SparseMatrix<E,symmetric>, SkewSymmetric > : identical<symmetric,SkewSymmetric> {};

template <typename E, typename symmetric>
class Rows< SparseMatrix<E,symmetric> >
   : public redirected_container< Rows< SparseMatrix<E,symmetric> >,
				  list( Container< sparse2d::Rows<E, symmetric::value, sparse2d::full, operations::masquerade2<sparse_matrix_line,symmetric> > >,
					MasqueradedTop ) > {
public:
   typename Rows::container& get_container()
   {
      return reinterpret_cast<typename Rows::container&>(*this->hidden().data);
   }
   const typename Rows::container& get_container() const
   {
      return reinterpret_cast<const typename Rows::container&>(*this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data->resize_rows(n);
   }
};

template <typename E, typename symmetric>
class Cols< SparseMatrix<E,symmetric> >
   : public redirected_container< Cols< SparseMatrix<E,symmetric> >,
				  list( Container< sparse2d::Cols<E, symmetric::value, sparse2d::full, operations::masquerade2<sparse_matrix_line,symmetric> > >,
					MasqueradedTop ) > {
public:
   typename Cols::container& get_container()
   {
      return reinterpret_cast<typename Cols::container&>(*this->hidden().data);
   }
   const typename Cols::container& get_container() const
   {
      return reinterpret_cast<const typename Cols::container&>(*this->hidden().data);
   }
   void resize(int n)
   {
      this->hidden().data->resize_cols(n);
   }
};

template <typename E, typename Permutation> inline
SparseMatrix<E,NonSymmetric> permuted_rows(const SparseMatrix<E,NonSymmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_rows - dimension mismatch");
   RestrictedSparseMatrix<E, sparse2d::only_rows> result(M.rows(), select(rows(M),perm).begin());
   return SparseMatrix<E,NonSymmetric>(result);
}

template <typename E, typename Permutation> inline
SparseMatrix<E,NonSymmetric> permuted_cols(const SparseMatrix<E,NonSymmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_cols - dimension mismatch");
   RestrictedSparseMatrix<E, sparse2d::only_cols> result(M.cols(), select(cols(M),perm).begin());
   return SparseMatrix<E,NonSymmetric>(result);
}

template <typename E, typename Permutation> inline
SparseMatrix<E,NonSymmetric> permuted_inv_rows(const SparseMatrix<E,NonSymmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_inv_rows - dimension mismatch");
   RestrictedSparseMatrix<E, sparse2d::only_rows> result(M.rows());
   std::copy(entire(rows(M)), select(rows(result),perm).begin());
   return SparseMatrix<E,NonSymmetric>(result);
}

template <typename E, typename Permutation> inline
SparseMatrix<E,NonSymmetric> permuted_inv_cols(const SparseMatrix<E,NonSymmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.cols() != perm.size(), "permuted_inv_cols - dimension mismatch");
   RestrictedSparseMatrix<E, sparse2d::only_cols> result(M.cols());
   std::copy(entire(cols(M)), select(cols(result),perm).begin());
   return SparseMatrix<E,NonSymmetric>(result);
}

template <typename E, typename symmetric, typename Permutation> inline
SparseMatrix<E,symmetric> permuted_rows(const SparseMatrix<E,symmetric>& M, const Permutation& perm)
{
   DIMENSION_CHECK(M.rows() != perm.size(), "permuted_rows - dimension mismatch");
   std::vector<int> inv_perm(M.rows());
   inverse_permutation(perm,inv_perm);
   return M.copy_permuted(perm,inv_perm);
}

template <typename E, typename symmetric, typename Permutation> inline
SparseMatrix<E,symmetric>
permuted_inv_rows(const SparseMatrix<E,symmetric>& M, const Permutation& inv_perm,
		  typename enable_if<int, derived_from<typename container_traits<Permutation>::category, random_access_iterator_tag>::value>::type=0)
{
   DIMENSION_CHECK(M.rows() != inv_perm.size(), "permuted_inv_rows - dimension mismatch");
   std::vector<int> perm(M.rows());
   inverse_permutation(inv_perm,perm);
   return M.copy_permuted(perm,inv_perm);
}

template <typename E, typename symmetric, typename Permutation> inline
SparseMatrix<E,symmetric>
permuted_inv_rows(const SparseMatrix<E,symmetric>& M, const Permutation& inv_perm,
		  typename disable_if<int, derived_from<typename container_traits<Permutation>::category, random_access_iterator_tag>::value>::type=0)
{
   DIMENSION_CHECK(M.rows() != inv_perm.size(), "permuted_inv_rows - dimension mismatch");
   std::vector<int> inv_perm_copy(inv_perm.size());
   std::copy(entire(inv_perm), inv_perm_copy.begin());
   return permuted_inv_rows(M,inv_perm_copy);
}

template <typename E, typename symmetric, typename Permutation> inline
SparseMatrix<E,symmetric> permuted_cols(const SparseMatrix<E,symmetric>& M, const Permutation& perm)
{
   return permuted_rows(M,perm);
}

template <typename E, typename symmetric, typename Permutation> inline
SparseMatrix<E,symmetric> permuted_inv_cols(const SparseMatrix<E,symmetric>& M, const Permutation& inv_perm)
{
   return permuted_inv_rows(M,inv_perm);
}

/// sparse matrix statistics collection
template <typename E>
class SparseMatrixStatistics {
public:
   unsigned int maxnon0s, maxrowsize, maxcolsize;
   E maxabs;

   SparseMatrixStatistics()
      : maxnon0s(0), maxrowsize(0), maxcolsize(0), maxabs(0) {}

   void gather(const SparseMatrix<E>& m)
   {
      unsigned int non0s=0;
      for (typename Entire< Rows< SparseMatrix<E> > >::const_iterator r=entire(rows(m)); !r.at_end(); ++r) {
	 if (unsigned int s=r->size()) {
	    for (typename Entire< typename SparseMatrix<E>::row_type >::const_iterator e=entire(*r); !e.at_end(); ++e)
	       maxabs=std::max(maxabs, abs(*e));
	    maxrowsize=std::max(maxrowsize, s);
	    non0s+=s;
	 }
      }
      maxnon0s=std::max(maxnon0s, non0s);
      for (typename Entire< Cols< SparseMatrix<E> > >::const_iterator c=entire(cols(m)); !c.at_end(); ++c) {
	 if (unsigned int s=c->size()) {
	    maxcolsize=std::max(maxcolsize, s);
	 }
      }
   }

   void gather(const Transposed< SparseMatrix<E> >& m)
   {
      gather(m.hidden());
   }

   // statistics at two various moments can be glued together
   SparseMatrixStatistics& operator+= (const SparseMatrixStatistics& s)
   {
      maxnon0s=std::max(maxnon0s,s.maxnon0s);
      maxabs=std::max(maxabs,s.maxabs);
      maxrowsize=std::max(maxrowsize,s.maxrowsize);
      maxcolsize=std::max(maxcolsize,s.maxcolsize);
      return *this;
   }

   template <typename Traits> friend
   std::basic_ostream<char, Traits>&
   operator<< (std::basic_ostream<char, Traits>& os, const SparseMatrixStatistics& s)
   {
      return os << ">>> " << s.maxnon0s << " nonzeroes,  max abs(element)=" << s.maxabs
		<< "\n>>> max row size=" << s.maxrowsize << ",  max col size=" << s.maxcolsize << endl;
   }
};

} // end namespace pm

namespace polymake {
   using pm::SparseMatrix;
   using pm::RestrictedSparseMatrix;
}

namespace std {
   template <typename E, typename symmetric> inline
   void swap(pm::SparseMatrix<E,symmetric>& M1, pm::SparseMatrix<E,symmetric>& M2)
   {
      M1.swap(M2);
   }

   template <typename E, pm::sparse2d::restriction_kind restriction> inline
   void swap(pm::RestrictedSparseMatrix<E,restriction>& M1,
	     pm::RestrictedSparseMatrix<E,restriction>& M2)
   {
      M1.swap(M2);
   }

   template <typename tree_type, typename symmetric> inline
   void swap(pm::sparse_matrix_line<tree_type,symmetric>& l1, pm::sparse_matrix_line<tree_type,symmetric>& l2)
   {
      l1.swap(l2);
   }
}

#endif // _POLYMAKE_SPARSE_MATRIX_H

// Local Variables:
// mode:C++
// End:
