/* Copyright (c) 1997-2008
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_SPARSE_VECTOR_H
#define _POLYMAKE_SPARSE_VECTOR_H "$Project: polymake $$Id: SparseVector.h 8018 2008-03-07 14:20:19Z gawrilow $"

#ifndef _POLYMAKE_GENERIC_VECTOR_H
#  include <GenericVector.h>
#endif
#ifndef _POLYMAKE_TREE_CONTAINERS_H
#  include <tree_containers.h>
#endif
#ifndef _POLYMAKE_SPARSE_H
#  include <sparse.h>
#endif
#ifndef _POLYMAKE_SHARED_OBJECT_H
#  include <shared_object.h>
#endif
#ifndef _POLYMAKE_AVL_H
#  include <AVL.h>
#endif

namespace pm {

template <typename NodeRef>
class sparse_vector_accessor : public AVL::node_accessor<NodeRef> {
   typedef AVL::node_accessor<NodeRef> _super;
public:
   typedef typename inherit_ref<typename deref<NodeRef>::type::mapped_type, typename _super::argument_type>::type result_type;
   result_type operator() (typename _super::argument_type n) const { return n.key_and_data.second; }
};

template <typename NodeRef>
struct sparse_vector_index_accessor {
   typedef NodeRef argument_type;
   typedef const int result_type;
   result_type operator() (NodeRef n) const { return n.key_and_data.first; }
};

template <typename E, typename Predicate>
class SparseVector
   : public modified_tree< SparseVector<E,Predicate>,
			   list( Container< AVL::tree< AVL::traits<int, E, operations::cmp> > >,
				 Operation< pair< BuildUnary<sparse_vector_accessor>,
				                  BuildUnary<sparse_vector_index_accessor> > > ) >,
     public GenericVector<SparseVector<E,Predicate>, E> {
   typedef modified_tree<SparseVector> _super;
protected:
   typedef AVL::tree< AVL::traits<int, E, operations::cmp> > tree_type;
   struct impl {
      tree_type tree;
      int d;
      Predicate non_zero;
 
      impl() : d(0) {}
      void clear() { d=0; tree.clear(); }
   };
   shared_object<impl, AliasHandler<shared_alias_handler> > data;

   friend SparseVector& make_mutable_alias(SparseVector& alias, SparseVector& owner)
   {
      alias.data.make_mutable_alias(owner.data);
      return alias;
   }

   typedef sparse_proxy_base<SparseVector> proxy_base;

   template <typename Iterator>
   void init(Iterator src, int dim_arg)
   {
      data.get()->d=dim_arg;
      data.get()->tree.assign(src);
   }

public:
   typedef random_access_iterator_tag container_category;
   typedef sparse_elem_proxy<proxy_base> reference;

   tree_type& get_container() { return data->tree; }
   const tree_type& get_container() const { return data->tree; }

   int dim() const { return data->d; }

   SparseVector() {}

   explicit SparseVector(int dim_arg) { data.get()->d=dim_arg; }

   template <typename Iterator>
   SparseVector(int dim_arg, Iterator src, typename enable_if_iterator<Iterator,end_sensitive>::type=0)
   {
      init(construct_sparse_iterator<Iterator,E>::_do(src, dim_arg), dim_arg);
   }

   /// Copy of a disguised SparseVector object.
   SparseVector(const GenericVector<SparseVector>& v)
      : data(v.top().data) {}

   template <typename Vector2>
   SparseVector(const GenericVector<Vector2,E>& v)
   {
      init(ensure(v.top(), (pure_sparse*)0).begin(), v.dim());
   }

   template <typename Vector2, typename E2>
   explicit SparseVector(const GenericVector<Vector2,E2>& v)
   {
      init(ensure(attach_converter<E>(v.top()), (pure_sparse*)0).begin(), v.dim());
   }

   template <int n>
   explicit SparseVector(const E (&a)[n])
   {
      init(ensure(array2container(a), (pure_sparse*)0).begin(), n);
   }

   template <typename E2, int n>
   explicit SparseVector(const E2 (&a)[n])
   {
      init(ensure(attach_converter<E>(array2container(a)), (pure_sparse*)0).begin(), n);
   }

   void clear() { data.apply(shared_clear()); }

   void resize(int n)
   {
      if (n<data->d) {
	 typename _super::reverse_iterator i=this->rbegin();
	 while (!i.at_end() && i.index()>=n) this->erase(i++);
      }
      data->d=n;
   }

   SparseVector& operator= (const SparseVector& v) { data=v.data; return *this; }

   using SparseVector::generic_type::operator=;

   void swap(SparseVector& v) { data.swap(v.data); }

   friend void relocate(SparseVector* from, SparseVector* to)
   {
      relocate(&from->data, &to->data);
   }

   reference operator[] (int i) { return proxy_base(*this,i); }

   const E& operator[] (int i) const
   {
      operations::clear<E> zero;
      return zero.deref_sparse(this->find(i));
   }

   template <typename Vector2>
   SparseVector& operator|= (const GenericVector<Vector2,E>& v)
   {
      const int d=data->d;
      tree_type& t=data->tree;
      for (typename ensure_features<Vector2,pure_sparse>::const_iterator src=ensure(v.top(), (pure_sparse*)0).begin();  !src.at_end();  ++src)
	 t.push_back(src.index()+d, *src);
      data->d += v.dim();
      return *this;
   }

   SparseVector& operator|= (const E& r)
   {
      if (data->non_zero(r)) push_back(data->d, r);
      data->d++;
      return *this;
   }

   template <typename Permutation> friend
   SparseVector permuted(const SparseVector& v, const Permutation& perm)
   {
      DIMENSION_CHECK(v.dim() != perm.size(), "permuted - dimension mismatch");
      SparseVector result(v.dim());
      tree_type& t=result.data.get()->tree;
      for (typename ensure_features<Permutation, cons<end_sensitive,indexed> >::const_iterator p=ensure(perm, (cons<end_sensitive,indexed>*)0);
	   !p.at_end();  ++p) {
	 typename SparseVector::const_iterator e=v.find(*p);
	 if (!e.at_end()) t.push_back(p.index(), *e);
      }
      return result;
   }

   template <typename Permutation> friend
   SparseVector permuted_inv(const SparseVector& v, const Permutation& perm)
   {
      DIMENSION_CHECK(v.dim() != perm.size(), "permuted_inv - dimension mismatch");
      SparseVector result(v.dim());
      tree_type& t=result.data.get()->tree;
      int pos=0;
      typename Permutation::const_iterator p=perm.begin();
      for (typename Entire<SparseVector>::const_iterator e=entire(v);  !e.at_end();  ++e) {
	 std::advance(p, e.index()-pos);
	 pos=e.index();
	 t.insert(*p,*e);
      }
      return result;
   }

#ifdef POLY_DEBUG
   void check() const { data->tree.check(""); }
   void tree_dump() const { data->tree.dump(); }
#endif

protected:
   friend class GenericVector<SparseVector>;

   void assign(const GenericVector<SparseVector>& v) { data=v.top().data; }

   template <typename Vector2>
   void assign(const GenericVector<Vector2>& v)
   {
      if (data.is_shared()) {
	 *this=SparseVector(v);
      } else {
	 data.get()->tree.assign(ensure(v.top(), (pure_sparse*)0).begin());
	 data.get()->d=v.dim();
      }
   }

   template <typename E2, int _size>
   void assign(const E2 (&a)[_size])
   {
      if (data.is_shared()) {
	 *this=SparseVector(a);
      } else {
	 data.get()->tree.assign(ensure(array2container(a), (pure_sparse*)0).begin());
	 data.get()->d=_size;
      }
   }

   template <typename Operation>
   void assign_op(const Operation& op)
   {
      if (data.is_shared())
	 *this=SparseVector(LazyVector1<const SparseVector&, Operation>(*this,op));
      else
	 SparseVector::generic_type::assign_op(op);
   }

   template <typename Vector2, typename Operation>
   void assign_op(const Vector2& v, const Operation& op)
   {
      if (data.is_shared())
	 *this=SparseVector(LazyVector2<const SparseVector&, const Vector2&, Operation>(*this,v,op));
      else
	 SparseVector::generic_type::assign_op(v,op);
   }

   template <typename E2>
   void _fill(const E2& x, pure_sparse)
   {
      data->tree.clear();
      if (data.get()->non_zero(x)) {
	 tree_type& t=data.get()->tree;
	 const int d=data.get()->d;
	 for (int i=0; i<d; ++i) t.push_back(i,x);
      }
   }
};

template <typename E, typename Predicate>
struct check_container_feature<SparseVector<E, Predicate>, pure_sparse> : True {};

} // end namespace pm

namespace polymake {
   using pm::SparseVector;
}

namespace std {
   template <typename E, typename Predicate> inline
   void swap(pm::SparseVector<E, Predicate>& v1, pm::SparseVector<E,Predicate>& v2) { v1.swap(v2); }
}

#endif // _POLYMAKE_SPARSE_VECTOR_H

// Local Variables:
// mode:C++
// End:
