#  Copyright (c) 1997-2009
#  Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
#  http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#----------------------------------------------------------------------------
#  $Project: polymake $$Id: InteractiveHelp.pm 9026 2009-01-09 18:03:47Z gawrilow $

use strict;
use namespaces;

#################################################################################
#
#  Help class designed for interactive shell
#
package Polymake::Core::InteractiveHelp;

use Polymake::Struct (
   [ new => ';$$$$' ],
   [ '$parent' => 'weak( #1 )' ],
   [ '$name' => '#2' ],
   '$category',
   [ '$text' => '#3' ],
   [ '$cloned' => '#4' ],
   '@toc',
   '%topics',
);

# between several entries with the same topic path
declare $separator="-----\n";

sub full_path {
   my $self=shift;
   my $path=$self->name;
   while (defined ($self=$self->parent)) {
      $path=$self->name."/$path";
   }
   $path
}

#################################################################################
my $stripped=qr{ [ \t]* (.*(?<!\s)) [ \t]*\n }xm;

sub add {
   my ($self, $path, $text, $split_multi)=@_;
   $path=[ split m'/', $path ] unless ref $path;
   if (is_object($path->[0])) {
      splice @$path, 0, 1, 'objects', $path->[0]->name;
   } elsif ($path->[0] eq 'objects' && $path->[1] =~ /::/ && UNIVERSAL::isa($path->[1], "Polymake::Core::Object")) {
      $path->[1] =~ s/^.*::($id_re)$/$1/o;
   }

   my ($cat, $author);
   if ($text =~ s/^\#?\s* category: $stripped//xomi) {
      splice @$path, -1, 0, $1;
      $cat=1;
   }
   if ($text =~ s/^\#?\s* author: $stripped//xomi) {
      $author=$1;
   }
   sanitize_help($text);
   return if !$cat && $text !~ /\S/;

   if (@$path) {
      my $topic=pop @$path;
      foreach (@$path) {
	 if (defined (my $h=$self->topics->{$_})) {
	    $self=$h;
	 } else {
	    push @{$self->toc}, $_;
	    $self=$self->topics->{$_}=new Help($self, $_);
	 }
      }
      $self->category ||= $cat;

      if (defined (my $h=$self->topics->{$topic})) {
	 if (!$split_multi) {
	    if (length($h->text)) {
	       $h->text .= $separator . $text if index($h->text,$text)<0;
	    } else {
	       $h->text=$text;
	    }
	    return $h;
	 }
	 if (exists $h->topics->{"$split_multi#0"}) {
	    $self=$h;
	    $topic="$split_multi#".scalar @{$self->toc};
	 } else {
	    $self=$self->topics->{$topic}=new Help($self, $topic);
	    $h->name="$split_multi#0";
	    $self->topics->{$h->name}=$h;
	    $topic="$split_multi#1";
	 }
      } else {
	 push @{$self->toc}, $topic;
      }

      $self=($self->topics->{$topic}=new Help($self, $topic, $text));

   } else {
      if (length($self->text)) {
	 $self->text .= $separator . $text;
      } else {
	 $self->text=$text;
      }
   }

   if ($author) {
      $self->topics->{author}=new Help($self, "author", $author);
   }
   return $self;
}
#################################################################################

my $continued_text=qr{ [ \t]* (.* \n             	  # rest of the first line
		       (?: (?:^\#?[ \t]*\n)*          	  # empty lines in between
			   ^\#?(?:[ ]*\t[ \t]*|[ ]{6,})   # indented with tabs and/or more than 5 spaces
		           (?!\w+:) \S.* \n )*		  # and without new keyword at the beginning
		       ) }x;

sub add_function {
   my ($self, $path, $help_text)=@_;
   my (%annex, %options);
   while ($help_text =~ s/^\#?\s* (return|signature|template_params): \s*(.*)$//xm) {
      $annex{$1}=$2;
   }
   while ($help_text =~ s/^\#?\s* args: $continued_text//xom) {
      my $text=$1;
      sanitize_help($text);
      $text =~ s/^([^\n]+)\n+$/$1/s;	# if the visible text consists of a single line, chop the eol's.
      $annex{args}=$text;
   }
   while ($help_text =~ s/^\#?\s* option: \s* ($id_re) $continued_text//xom) {
      my ($name, $text)=($1, $2);
      sanitize_help($text);
      $options{$name}=new Help(undef, $name, $text);
      $annex{options}="";
   }

   if (my $h=$self->add($path, $help_text, exists $annex{signature} && "overload")) {
      while (my ($annex, $text)=each %annex) {
	 $h->topics->{$annex}=new Help($h, $annex, $text);
      }
      if (%options) {
	 $h->topics->{options}->topics=\%options;
      }
      $h
   }
}
#################################################################################
sub clone {
   my ($src, $up)=@_;
   my $self=new Help($up, $src->name, $src->text, 1);
   @{$self->toc}=@{$src->toc};
   while (my ($name, $topic)=each %{$src->topics}) {
      $self->topics->{$name}=$topic->clone($self);
   }
   $self->category=$src->category;
   $self
}
#################################################################################
sub merge {
   my ($self, $src)=@_;
   if (length($src->text)) {
      if (length($self->text)) {
	 $self->text .= $separator . $src->text;
      } else {
	 $self->text=$src->text;
      }
   }
   my %added;
   while (my ($name, $src_topic)=each %{$src->topics}) {
      if (defined (my $topic=$self->topics->{$name})) {
	 merge($topic, $src->topics->{$name});
      } else {
	 $self->topics->{$name}=$src_topic->clone($self);
	 $added{$name}=1;
      }
   }
   push @{$self->toc}, grep { $added{$_} } @{$src->toc};
   $self->category|=$src->category;
}
#################################################################################
sub get_topic {
   my $self=shift;
   my $path= @_ > 1 ? \@_ : is_ARRAY($_[0]) ? shift  : [ split m'/', shift ];
   shift @$path if @$path>1 && !length($path->[0]);
   foreach (@$path) {
      $self=$self->topics->{$_} or last;
   }
   $self
}
#################################################################################
sub apropos {
   my ($self, $expr)=@_;
   my @list;
   push @list, $self if $self->name =~ $expr;
   foreach my $subtopic (values %{$self->topics}) {
      push @list, $subtopic->apropos($expr);
   }
   @list
}
#################################################################################
sub get {
   my $self=&get_topic;
   $self && $self->cleaned_text;
}
#################################################################################
sub cleaned_text {
   my $self=shift;
   my $text=$self->text;
   $text =~ s/\@(?:[abci]|em)\s+//g;
   $text =~ s/\@see\s+$hier_id_re\s*\{($balanced_re)\}/$1/go;
   $text =~ s/\@see\s+//g;
   $text;
}
#################################################################################
# flags: 1: dive thru categories
#        2: own topics only
sub list_completions {
   my ($self, $prefix, $flags)=@_;
   map { ($flags & 1 & $_->category)
	 ? list_completions($_,$prefix,$flags&2) :
	 !(($flags&2) && $_->cloned) && index($_->name,$prefix)==0
	 ? $_->name : ()
   } values %{$self->topics}
}

sub find_in_categories {
   my ($self, $name)=@_;
   $self->topics->{$name} || do {
      my $subtopic;
      foreach my $cat (grep { $_->category & 1 } values %{$self->topics}) {
	 defined ($subtopic=$cat->topics->{$name}) and last;
      }
      $subtopic
   }
}

sub expects_template_params {
   my $self=shift;
   my $cnt=0;
   if (defined (my $template_params=$self->topics->{template_params}) and
       defined (my $signature=$self->topics->{signature})) {
      while ($template_params->text =~ /($id_re)/go) {
	 my $name=$1;
	 if ($signature->text =~ /\b$name\b/) {
	    last;
	 } else {
	    ++$cnt;
	 }
      }
   }
   $cnt
}

1
