#  Copyright (c) 1997-2009
#  Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
#  http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#----------------------------------------------------------------------------
#  $Project: polymake $$Id: Property.pm 9089 2009-02-19 17:37:14Z gawrilow $

use strict;
use namespaces;

package Polymake::Core::Property::_defs;
# for rulefile parser

sub canonical;
sub equal;

####################################################################################
package Polymake::Core::Property;

#  Constructor
#
#  new Property(ObjectType, 'Name', $type, $decl_args);

use Polymake::Struct (
   [ 'new' => '$$$$$' ],
   [ '$name' => '#2' ],		# 'property name'
   [ '$belongs_to' => '#1' ],	# ObjectType
   [ '$extension' => '$Application::extension' ],
   [ '$type' => '#3' ],		# PropertyType | ObjectType
   [ '$flags' => '0' ],
				# the following fields contain \&sub or "method name",
				#   methods are called via the Object
   '&canonical',		# sub: data =>
   '&equal',

   '&accept',			# value, object, trusted_flag =>
   [ '$inst_cache' => 'undef' ],
   [ '$up' => '0' ],		# how many parent objects to travel upwards in order to get a proper type context

   '%prod_keys',		# for producing rules in parent objects: tree of hashes: SubobjectProperty => { ... }
				# each sub-hash serves as key in ObjectType::producers
   '@aliases',			# alias names (currently kept here solely for documentation purposes)
   '$key',
);

# Property->flags
use Polymake::Enum qw( is: mutable=1 locally_extended=2 multiple=4 produced_as_whole=8 permutation=16 );

# Constructor: ObjectType, 'name', Type, attributes, open_scope =>

sub new {
   my $self=&_new;
   my $with_scope=pop;
   my $attrs=pop;
   if ($attrs =~ s/:\s*mutable\b//) {
      $self->flags|=$is_mutable;
   }
   if ($attrs =~ s/:\s*multiple\b//) {
      if ($Application::plausibility_checks && !instanceof ObjectType($self->type)) {
	 croak( "an atomic property may not be declared as multiple" );
      }
      $self->flags|=$is_multiple;
   }
   if ($attrs =~ /\S/) {
      croak( "unknown property attributes $attrs" );
   }
   if ($self->type->abstract) {
      $self->inst_cache={ };
      if (defined $self->type->context_pkg) {
	 my $obj_proto=$self->belongs_to;
	 while ($obj_proto->pkg ne $self->type->context_pkg  and  ! defined($obj_proto->descend_to_generic($self->type->context_pkg))) {
	    $obj_proto=$obj_proto->parent_property->belongs_to;
	    $self->up++;
	 }
      }
      $self->accept=sub : method {
	 my $self=shift;
	 ($self->clone($_[1]))->accept->(@_);
      };
   } else {
      $self->accept=choose_accept_method($self);
   }
   weak($self->key=$self);

   $self;
}
####################################################################################
sub clone {
   my ($src, $obj)=@_;
   my $up=$src->up;
   while ($up>0) { $obj=$obj->parent; --$up }
   my $obj_proto=$obj->type;
   $src->inst_cache->{$obj_proto} ||= do {
      my $self=inherit_class([ @$src ], $src);
      $self->type=$src->type->concrete_type($obj_proto);
      $self->accept=choose_accept_method($self);
      $self->key=$src;
      $self;
   }
}

sub concrete {
   my ($self, $obj)=@_;
   $self->type->abstract ? $self->clone($obj) : $self
}
####################################################################################
sub choose_accept_method {
   my ($self)=@_;
   instanceof ObjectType($self->type)
   ? \&accept_subobject :
   $self->type->cppoptions && !$self->type->cppoptions->default_constructor
   ? \&accept_special_constructed :
   $self->type->construct
   ? \&accept_composite
   : \&accept_atomic;
}
####################################################################################
sub check_subobject {
   my ($self, $value, $expected)=@_;
   if (!is_object($value) || !UNIVERSAL::isa($value, $expected->pkg)) {
      die "wrong value for a subobject property ", $self->name, ": ", ref($value) || "'$value'", "\n";
   }
}
####################################################################################
sub finalize_subobject {
   my ($self, $value, $obj, $trusted, $temp)=@_;
   if (defined($value->parent) || defined($value->persistent)) {
      if ($value->type==$self->type) {
	 $value=$value->copy(undef);
      } elsif ($self->flags & $is_locally_extended) {
	 # create the autonomous external type, get rid of potentially dangerous foreign extensions
	 $value=$self->type->super->[1]->construct->($value);
      } else {
	 $value=$self->type->construct->($value);
      }
      defined(wantarray) or $_[1]=$value;	# replace the subobject in the caller
   }
   if ($self->flags & $is_locally_extended) {
      my $orig_type=$value->type;
      if ($orig_type != $self->type) {
	 bless $value, $self->type->local_type($orig_type)->pkg;
      }
   }
   weak($value->parent=$obj);
   $value->property=$self;
   $value->is_temporary= $temp || defined($obj->parent) && $obj->is_temporary;
   if (defined($value->transaction)) {
      if ($value->has_cleanup) {
	 Transaction::merge_temporaries($value->transaction->temporaries, delete $Scope->cleanup->{$value});
	 $value->has_cleanup=0;
      }
      if (defined($obj->transaction)) {
	 $obj->transaction->descend($value) unless ($temp && defined($obj->transaction->rule));
      } else {
	 $value->transaction->commit($value);
      }
   } elsif (defined($obj->transaction)) {
      $obj->transaction->descend($value) unless ($temp && defined($obj->transaction->rule));
      if ($value->has_cleanup) {
	 $value->transaction->temporaries=delete $Scope->cleanup->{$value};
	 $value->has_cleanup=0;
      }
   }
   if (!$trusted && defined($self->canonical)) {
      select_method($self->canonical,$obj,1)->($value);
   }
   defined(wantarray) and $self->flags & $is_multiple ? new PropertyMultipleValue($self, [ $value ]) : $value;
}
####################################################################################
sub accept_subobject : method {
   my ($self, $value, $obj, $trusted)=@_;
   if (defined $value) {
      $trusted or check_subobject($self, $value, $self->flags & $is_locally_extended ? $self->type->super->[1] : $self->type);
      &finalize_subobject;
   } else {
      if (!$trusted and $self->flags & $is_multiple) {
	 croak( "undefined value not allowed for the multiple subobject property ", $self->name );
      }
      new PropertyValue($self, $value);
   }
}
####################################################################################
sub accept_composite : method {
   my ($self, $value, $obj, $trusted, $temp)=@_;
   if (defined $value) {
      my $needs_canonicalization=!$trusted && defined($self->canonical);
      if (is_string($value)) {
	 local $PropertyType::trusted_value=$trusted;
	 $value=$self->type->parse_string->($value);
      } elsif (!is_object($value) or
	       !$self->type->isa->($value) or
	       $self->type->cppoptions && !$self->type->cppoptions->builtin &&
	       CPlusPlus::must_be_copied($value,$temp,$needs_canonicalization)) {
	 local $PropertyType::trusted_value=$trusted;
	 $value=$self->type->construct->($value);
      }
      if ($needs_canonicalization) {
	 select_method($self->canonical,$obj,1)->($value);
      }
   }
   new PropertyValue($self, $value, $temp);
}
####################################################################################
sub accept_special_constructed : method {
   my ($self, $value, $obj, $trusted, $temp)=@_;
   if (defined($value)) {
      my $needs_canonicalization=!$trusted && defined($self->canonical);
      if (!is_object($value) or
	  !$self->type->isa->($value) or
	  CPlusPlus::must_be_copied($value,$temp,$needs_canonicalization)) {
	 local $PropertyType::trusted_value=$trusted;
	 $value=$self->type->construct->($obj,$value);
      }
      if ($needs_canonicalization) {
	 select_method($self->canonical,$obj,1)->($value);
      }
   }
   new PropertyValue($self, $value, $temp);
}
####################################################################################
sub accept_atomic : method {
   my ($self, $value, $obj, $trusted, $temp)=@_;
   if (defined $value) {
      if (is_string($value)) {
	 $value=$self->type->parse_whole($value, $trusted);
      }
      if (!$trusted && defined($self->canonical)) {
	 select_method($self->canonical,$obj,1)->($value);
      }
   }
   new PropertyValue($self, $value, $temp);
}
####################################################################################
# ancestor_prop, ... => last key
# arguments come in reverse (bottom-up) order!
sub get_prod_key {
   my $self=shift;
   my $hash=$self->prod_keys;
   my $perm_seen;
   foreach (@_) {
      $perm_seen ||= $_->flags & $Property::is_permutation;
      $hash=($hash->{$_} ||= new SubobjKey($self,$_,$perm_seen));
   }
   $hash
}

package Polymake::Core::Property::SubobjKey;
RefHash::allow(__PACKAGE__);

my $prop_key=\(1);
my $subobj_key=\(2);
my $perm_key=\(3);

sub new {
   my $self=bless { $prop_key => $_[1], $subobj_key => $_[2] }, $_[0];
   weak($_) for values %$self;
   $self->{$perm_key}=1 if $_[3];
   $self;
}

sub property { (shift)->{$prop_key} }
sub subobject_property { (shift)->{$subobj_key} }
sub belongs_to { (&subobject_property)->belongs_to }
sub on_permutation_path { (shift)->{$perm_key} }

####################################################################################
package Polymake::Core::Property::Finalizer;

use Polymake::Struct (
   [ new => '$' ],
   [ '$property' => '#1' ],
   [ '$pkg' => 'undef' ],
);

sub import {
   my ($self, $pkg)=@_;
   $self->pkg=$pkg;
   if (instanceof ObjectType($self->property->type)) {
      new ExpectingLocalExtension($self);
   } else {
      namespaces::using($pkg, 'Polymake::Core::Property::_defs', $self->property->type->pkg);
   }
}

sub DESTROY {
   unless ($@) {
      PropertyType::Finalizer::store_methods(@{$_[0]}, qw(canonical equal));
   }
}

1
